

namespace eval ::main {
    set devilVersion "0.16.0"
    set initProc ""
    set appHome [file dirname [file join [pwd] $::argv0]]
}


# --- public functions callable from other modules -------------------


proc main::getRessourcePath {relativePath} {
  return [file join $main::appHome $relativePath]
}


proc main::getTempPath {fileName} {
  if { [info exists ::env(TEMP)] } {
    set tempDir $::env(TEMP)
  } else {
    set tempDir "/tmp"
  }
  return [file join $tempDir $fileName]
}


proc main::addInitProc {cmd} {
    lappend main::initProc $cmd
}


proc main::getValidImage {name} {
    if { [catch {image type $name}] } {
        log::log error "Unknown image: '$name'"
        return "img::default"
    } else {
        return $name
    }
}


proc main::getParam {key} {
    return $::main::param($key)
}


proc main::loadSpec {file} {
    loadTcl spec $file
}


proc main::loadLib {file} {
    loadTcl lib $file
}



# --- loading modules -------------------------------------------------


proc main::loadTcl {level file} {
    # Loads $file in $module (if not already loaded).
    
    variable loaded
    switch $level {
        lib {
            set filename "$main::appHome/vlEnv/$file"
        }
        spec {
            set filename "$main::appHome/otherTcl/$file"
        }
        default {
            error "Invalid level '$level'. Must be 'lib' or 'spec'"
        }
    }
    if { ![info exists loaded($filename)] } {
        set loaded($filename) 1
        namespace eval :: [list source $filename]
    }
}



# --- splash screen -------------------------------------------------------


proc main::loadSplash {} {

  variable splashExistent
  set splashExistent 0

  if { [lsearch -exact [image names] "img::splash"] != -1 } {
     toplevel .splashWindow
     wm title .splashWindow "DEViL"
     wm overrideredirect .splashWindow 1
     label .splashWindow.splashImage -text "Image" -image "img::splash"
     pack .splashWindow.splashImage -side left
     set splashExistent 1
   }
}

proc main::centerWindow {w} {
  wm withdraw $w
  update idletasks
  set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 - [winfo vrootx [winfo parent $w]]]
   set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 - [winfo vrooty [winfo parent $w]]]
  wm geom $w +$x+$y
  wm deiconify $w
}


proc main::unloadSplash {} {
    variable splashExistent
    if { $splashExistent } {
        centerWindow .splashWindow
        raise .splashWindow
        after 4000 {destroy .splashWindow}
    }
}

# --- parameter processing -----------------------------------------------------------------------------


proc main::error {msg} {
    puts $msg
    exit 1
}

proc main::showCommandLineHelp {} {
    puts "Syntax: $cfg::commandName \[options\] \[--\] \[file1 file2 ...\]"
    puts "        $cfg::commandName -h"
    puts ""
    puts "Options:"
    puts "  -debug 1                   Activates debug mode."
    puts "  -process <path>            Process the given file."
    puts "  -processor <processorName> Use the given processor to process a file."
    puts "                             <processorName> may also be 'EXPORT_HTML'."
    puts "  -dest <path>               Store the result of file processing at given location."
    puts ""
}


proc main::parseParams_helper {} {

    # default values
    set ::main::param(files) ""
    set ::main::param(debug) 0
    set ::main::param(process) ""
    set ::main::param(processor) ""
    set ::main::param(dest) ""
    
    set state optionOrFile

    set eolMsg ""
    foreach param $::argv {
        set eolMsg ""
        switch $state {
            optionOrFile {
                if { $param eq "--" } {
                    set state file
                } elseif { $param eq "-h" } {
                    showCommandLineHelp
                    exit
                } elseif { [regexp "^-(.*)" $param dummy option] } {
                    if { [lsearch -exact "debug process processor dest" $option] == -1 } {
                        return "Unknown option $option"
                    }
                    set eolMsg "Option value missing"
                    set state optionValue
                } else {
                    lappend ::main::param(files) $param
                    set state file
                }
            }
            optionValue {
                set ::main::param($option) $param
                set state optionOrFile
            }
            file {
                lappend ::main::param(files) $param
                set state file                
            }
        }
    }
    return $eolMsg
}


proc main::parseParams {} {
    set msg [parseParams_helper]
    if { $msg ne "" } {
        puts $msg
        puts "Enter '$cfg::commandName -h' for help"
        exit
    }
}


proc main::autoProcess_helper {fileId} {
    if { [file::startProcessor $fileId [getParam processor]] } {
        exit 0
    }
}

proc main::autoProcess {fileId} {
    after idle [list main::autoProcess_helper $fileId]
    # BOGUS Komischer effekt: ohne diese Indirektion
    # oeffnet sich im Fehlerfall die Oberflaeche nicht.
}


proc main::processParams {} {
    foreach option "debug process processor dest files" {
        log::log options "$option: [getParam $option]"
    }
    if { [getParam process] ne "" } {
        set fileId [file::openNoWindow [getParam process]]
        file::afterUpdate $fileId [list main::autoProcess $fileId]
        return
    } else {
        set log::mdiOutput 1
        update idletasks
        # xxx-id
    }
    foreach file [getParam files] {
        file::openWithWindow $file
    }
    set log::mdiOutput 1
}



# --- init function ----------------------------------------------------------------


proc main::main {} {

    #### FIRST, parse parameters.

    parseParams
    # Parameter parsing must be the first step, because
    # the -debug options determines, whether debug code in the function
    # definitions is enabled or disabled.


    #### NEXT, load all modules and resources

    # Init c-modules
    Class::reinit

    # Load images
    foreach file [glob "$main::appHome/img/*.gif"] {
        set name img::[file tail [file rootname $file]]
        image create photo $name -file $file
    }

    # Load libraries
    loadLib "xn.tcl"
    ::xn::setDebug [getParam debug]

    loadLib "log.tcl"
    foreach file [lsort [glob "$main::appHome/vlEnv/*.tcl"]] {
        loadLib [file tail $file]
    }

    # Load mdi
    namespace eval :: [list source "$main::appHome/mdi/mdi.tcl"]

    # Load specs
    if { [file exists "$main::appHome/otherTcl/meta.tcl"] } {
        loadSpec meta.tcl
    }
    foreach file [lsort [glob "$main::appHome/otherTcl/*.tcl"]] {
        loadSpec [file tail $file]
    }

    #### NEXT, initialize all modules

    loadSplash
    i18n::init
    MDIFrame::Setup

    # Call init procs
    foreach proc $::main::initProc {
       namespace eval :: $proc
    }
    processParams
    unloadSplash
}


main::main

