

namespace eval ::db {}

set db::driverClass(mysql) "com.mysql.jdbc.Driver"
set db::driverClass(firebird) "org.firebirdsql.jdbc.FBDriver"
set db::driverClass(oracle) "oracle.jdbc.driver.OracleDriver"
set db::driverClass(hsqldb) "org.hsqldb.jdbcDriver"

set db::dialect(mysql) "org.hibernate.dialect.MySQLDialect"
set db::dialect(firebird) "org.hibernate.dialect.FirebirdDialect"
set db::dialect(oracle) "org.hibernate.dialect.OracleDialect"
set db::dialect(hsqldb) "org.hibernate.dialect.HSQLDialect"

proc db::getDriverClass {dbtype} {

  if ![info exists db::driverClass($dbtype)] {
    return "$dbtype: unknown database!"
  }
  
  return $db::driverClass($dbtype)
}

proc db::getDialect {dbtype} {

  if ![info exists db::dialect($dbtype)] {
    return "$dbtype: unknown database!"
  }
  
  return $db::dialect($dbtype)
}

proc db::getUrl {dbtype dbname host port} {

  if [equal $dbtype "mysql"] {
    return "jdbc:mysql://$host:$port/$dbname"
  } elseif [equal $dbtype "firebird"] {
    return "jdbc:firebirdsql:$host/$port:$dbname"
  } elseif [equal $dbtype "hsqldb"] {
    return "jdbc:hsqldb:hsql://$host:$port/$dbname"
  } elseif [equal $dbtype "oracle"] {
    return "jdbc:oracle:thin:@$host:$port:$dbname"
  }
  
  return "$dbtype: unknown database!"
}

proc traverseConditionTree {objId {extraCondition ""}} {

 if {[nullptr $objId] || [equal $objId ""]} {
   return $extraCondition
 }
 
 set className [c::getClassName $objId]
 
 if [ equal $className "BinaryExpression" ] {
  set left [c::getList "$objId.leftOperand.CHILDREN"]
  set right [c::getList "$objId.rightOperand.CHILDREN"]
  set operator [c::get "$objId.operator.VALUE"]
  if ![nullptr $left] {
   if ![nullptr $right] {
     if ![equal $extraCondition ""] {
       return [ concat "( $extraCondition and ("  [traverseConditionTree $left] $operator [traverseConditionTree $right] "))" ]
     }
     return [ concat "("  [traverseConditionTree $left] $operator [traverseConditionTree $right] ")" ]
   }
  }
   return ""
 }

 if [ equal $className "UnaryExpression" ] {
  set operand [c::getList "$objId.operand.CHILDREN"]
  set operator [c::get "$objId.operator.VALUE"]
  if [equal $operator "not"] {
    if ![equal $extraCondition ""] {
      return [ concat "$extraCondition and (not " [traverseConditionTree $operand]  ")" ]
    }
    return [ concat "(not " [traverseConditionTree $operand]  ")" ]
  }
  
  if ![equal $extraCondition ""] {
    return [ concat "$extraCondition and (" [traverseConditionTree $operand] $operator  ")" ]
  }
  return [ concat "(" [traverseConditionTree $operand] $operator  ")" ]
 }
 
 if [ equal $className "PredefinedExpression" ] {
  return [c::get "$objId.value.VALUE"]
 }
 
 if [equal $className "ColumnValueExpression" ] {
      set column [c::get "$objId.column.VALUE.name.VALUE"]
      set columnObj [c::get "$objId.column.VALUE"]
      set dataSource [c::get "$columnObj.INCLUDING DataSourceTable.name.VALUE"]
   if ![nullptr $column] {
    return [string tolower "$dataSource.$column"]
    } {
      return ""
    }
 }
 
 if [equal $className "ProjectionColumnValueExpression" ] {
      set alias [c::get "$objId.column.VALUE.name.VALUE"]
      set columnObj [c::get "$objId.column.VALUE.column.VALUE"]
      set dataSource [c::get "$columnObj.INCLUDING DataSourceTable.name.VALUE"]
   if ![nullptr $alias] {
     return [string tolower "$dataSource.$alias"]
   } {
     return ""
   }
 }
 
 if [equal $className "ConstantExpression"] {
   return [getConstantExpressionValue $objId "'"]
 }

 if [equal $className "ParameterValueExpression"] {
   set formalParameter [c::get "$objId.parameter.VALUE"]
   if [nullptr $formalParameter] {
     return ""
   }
   set paramName [c::get "$formalParameter.name.VALUE"]
   return ":$paramName" 
 }

 if [equal $className "ComputationExpression"] {
   set column [c::get "$objId.column.VALUE.name.VALUE"]
   set columnObj [c::get "$objId.column.VALUE"]
   if ![nullptr $column] {
     set computation [c::get "$objId.computation.VALUE"]
     return [transformComputation $column $computation [string tolower $computation] $columnObj]
    } {
      return ""
    }
 }
  return ""
}

proc transformComputation {column computation alias columnObj} {
  set dataSource  [string tolower [c::get "$columnObj.INCLUDING DataSourceTable.name.VALUE"]]
  return [format "%s%s" $computation "($dataSource.$column) as $alias"]
}

proc getDataSourceSelectHQL {dsselect} {
  set Ds [string totitle [c::get $dsselect.dataSource.VALUE.name.VALUE]]
  set ds [string tolower $Ds]
  set value [string tolower [c::get $dsselect.value.VALUE.name.VALUE]]
  set label [string tolower [c::get $dsselect.label.VALUE.name.VALUE]]
  return "select $ds.$value, $ds.$label from $Ds $ds"
}

proc getUpdateHQLQuery {objId} {
  
  set valuesList [c::getList "$objId.values.CHILDREN"]
  set valuesLength [llength $valuesList]
  set values  ""
  set counter 1
  
  set ds [string tolower [getTargetDataSourceName $objId]]
  set Ds [string totitle $ds]
  
  foreach {listEntry} $valuesList {
    set expression [c::getList "$listEntry.expression.CHILDREN"]
    set class [c::getClassName $expression]
    set columnObj [c::get "$listEntry.column.VALUE"]
    set columnRef [string tolower [c::get "$columnObj.name.VALUE"]]
    set columnType [c::getList "$columnObj.type.CHILDREN"]
    set columnTypeClass [c::getClassName $columnType]
   
    if [equal $class "ConstantExpression"] {
      set values [concat $values "$ds.$columnRef =" [getConstantExpressionValue $expression "'"] ] 
    } elseif [equal $class "ParameterValueExpression"] {
      set formalParameter [c::get "$expression.parameter.VALUE"]
      if [nullptr $formalParameter] {
        continue
      }
      set paramType [common::getTypeClassName $formalParameter]
      set paramName [c::get "$formalParameter.name.VALUE"]
      if [equal $paramType "DBBlob"] {
        set values [concat $values [format "%s%s%s" "$ds.$columnRef" "ContentType = :$paramName" "_contenttype, "]]
        set values [concat $values [format "%s%s%s" "$ds.$columnRef" "Path = :$paramName" "_path"]]
      } {
        set values [concat $values "$ds.$columnRef = :$paramName"]
      }
    }
    
    if ![equal $valuesLength $counter] {
      set values [concat $values ", "]
    }
    set counter [expr $counter + 1]
  }
  set condition [lindex [c::getList "$objId.conditions.CHILDREN"] 0]
  return [concat "update $Ds $ds set" $values "where " [traverseConditionTree $condition]]
}

proc isLimitDataSource {objId} {
 set dataSource [c::get "$objId.dataSource.VALUE"]
 if ![nullptr $dataSource] {
   set className [c::getClassName "$dataSource"]
  if [equal $className "LimitDataSource"] {
   return true
  } {
   return false
  }
 }
 return false
}

proc getJoinCondition {startDataSource} {
 set className [c::getClassName $startDataSource]
 
 if [equal $className "DataSourceJoin"] {
   set dataSource1 [c::get "$startDataSource.dataSource1.VALUE"]
   set dataSource2 [c::get "$startDataSource.dataSource2.VALUE"]

   if ![nullptr $dataSource1] {
     if ![nullptr $dataSource2] {

       set d1className [c::getClassName $dataSource1]
       set d2className [c::getClassName $dataSource2]

        if [equal $d1className "DataSourceTable"] {
          if [equal $d2className "DataSourceTable"] {
           set columnRef1 [c::get "$startDataSource.columnRef1.VALUE"]
           set columnRef1Name [c::get "$startDataSource.columnRef1.VALUE.name.VALUE"]
           set columnRef1Table [c::get "$columnRef1.INCLUDING DataSourceTable.name.VALUE"]
           set columnRef2 [c::get "$startDataSource.columnRef2.VALUE"]
           set columnRef2Name [c::get "$startDataSource.columnRef2.VALUE.name.VALUE"]
           set columnRef2Table [c::get "$columnRef2.INCLUDING DataSourceTable.name.VALUE"]
           return [string tolower "$columnRef1Table.$columnRef1Name = $columnRef2Table.$columnRef2Name"]
          }
        }
     }
   }


 }

 return ""

}

proc getJoinTables {startDataSource} {
 set className [c::getClassName $startDataSource]
 
 if [equal $className "DataSourceJoin"] {
   set dataSource1 [c::get "$startDataSource.dataSource1.VALUE"]
   set dataSource2 [c::get "$startDataSource.dataSource2.VALUE"]

   if ![nullptr $dataSource1] {
     if ![nullptr $dataSource2] {

       set d1className [c::getClassName $dataSource1]
       set d2className [c::getClassName $dataSource2]
       
        if {[equal $d1className "DataSourceTable"] && [equal $d2className "DataSourceTable"]} {
          return [concat [getHQLFromTable $dataSource1] "," [getHQLFromTable $dataSource2]]
        }
     }
   }
 } elseif [equal $className "DataSourceTable"] {
   return [getHQLFromTable $startDataSource]
 }

 return ""
}

proc getHQLFromTable {dataSourceTable} {
  set dsName [string tolower [c::get "$dataSourceTable.name.VALUE"]]
  set DsName [string totitle $dsName]
  return [concat $DsName $dsName]
}

proc getSelectHQLQuery {objId} {
  ## Wir haben einen DataSourceProjection
  set condition [c::getList "$objId.condition.CHILDREN"]
  set projectionColumns [c::getList "$objId.projectionColumns.CHILDREN"]
  set projLength [llength $projectionColumns]
  
  set counter 1
  set projectionString ""
  
  foreach {listEntry} $projectionColumns {
    set class [c::getClassName $listEntry]
    set columnObj [c::get "$listEntry.column.VALUE"]
    set columnName [string tolower [c::get "$columnObj.name.VALUE"]]
    set columnTypeName [common::getTypeClassName $columnObj]
    set alias [string tolower [c::get "$listEntry.name.VALUE"]]
        
    if [equal $class "ColumnProjection"] {
      set connectedDataSource [string tolower [c::get "$columnObj.INCLUDING DataSourceTable.name.VALUE"]]
      if [equal $columnTypeName "DBBlob" ] {
        set projectionString [concat $projectionString [format "%s%s%s" "$connectedDataSource.$columnName" "Path as $alias" "Path, "]]
        set projectionString [concat $projectionString [format "%s%s%s" "$connectedDataSource.$columnName" "ContentType as $alias" "ContentType"]]
      } {
        set projectionString [concat $projectionString "$connectedDataSource.$columnName as $alias"]
      }
    } elseif [equal $class "ComputationProjection"] {
      
      set transformation [transformComputation $columnName [c::get "$listEntry.computation.VALUE"] $alias $columnObj]
      set projectionString [concat $projectionString $transformation]

    }
    
    if ![equal $projLength $counter] {
      set projectionString [concat $projectionString ", "]
    }

    set counter [expr $counter + 1]

  }
  
  set joinTables "N/A"
  set joinCondition ""
  if ![nullptr [c::get "$objId.sourceDataSource"]] {
    set joinTables [getJoinTables [c::get "$objId.sourceDataSource.VALUE"]]
    set joinCondition [getJoinCondition [c::get "$objId.sourceDataSource.VALUE"]]    
  }
  
  set whereClause [traverseConditionTree $condition $joinCondition]
  if ![equal $whereClause ""] {
    set whereClause [concat "where" $whereClause]
  }
  return [concat "select new map(" $projectionString ") from" $joinTables $whereClause]
  
}

# Checks whether the given datasource is a join query
proc isJoinQuery {dataSource} {
  set dataSourceName [c::getClassName $dataSource]
  
  if [equal $dataSourceName "DataSourceJoin"] {
    return true
    
  } elseif [equal $dataSourceName "DataSourceTable"] {
    return false
    
  } elseif {[equal $dataSourceName "DataSourceProjection"] || [equal $dataSourceName "OrderedDataSource"] || [equal $dataSourceName "LimitDataSource"]} {
    return [isJoinQuery [c::get "$dataSource.sourceDataSource.VALUE"]]
    
  }
  
  return false
}

proc isProjectionDataSource {dataSource} {
  set dataSourceName [c::getClassName $dataSource]
  
  if {[equal $dataSourceName "DataSourceJoin"] || [equal $dataSourceName "DataSourceTable"]} {
    return false
    
  } elseif {[equal $dataSourceName "OrderedDataSource"] || [equal $dataSourceName "LimitDataSource"]} {
    return [isProjectionDataSource [c::get "$dataSource.sourceDataSource.VALUE"]]
    
  } elseif [equal $dataSourceName "DataSourceProjection"] {
    return true
  }
  
  return false
}

proc getHQLQuery {dataSource} {
  set dataSourceName [c::getClassName $dataSource]
  
  if [equal $dataSourceName "DataSourceTable"] {
    return [getDbHQLQuery $dataSource]
    
  } elseif [equal $dataSourceName "OrderedDataSource"] {
    return [getOrderedHQLQuery $dataSource]
    
  } elseif [equal $dataSourceName "LimitDataSource"] {
    return [getHQLQuery [c::get "$dataSource.sourceDataSource.VALUE"]]
    
  } elseif [equal $dataSourceName "DataSourceProjection"] {
    return [getSelectHQLQuery $dataSource]
    
  }
  
  return ""
}


proc getDbHQLQuery {dataSourceTable} {
  return [concat "from" [getHQLFromTable $dataSourceTable]]
}

proc getOrderedHQLQuery {objId} {
  set sourceDataSource [c::get "$objId.sourceDataSource.VALUE"]
  set orderColumns [c::getList $objId.orderColumns.CHILDREN]
  if [equal [llength $orderColumns] 0] {
   return [getHQLQuery $sourceDataSource]
  }
  
  set orderColumnsString ""
  set counter 1
  
  foreach {ocolumn} $orderColumns {
    set column [c::get $ocolumn.orderColumn.VALUE]
    set order [c::get $ocolumn.order.VALUE]
    set columnName [string tolower [c::get $column.name.VALUE]]
    set ds [string tolower [c::get "$column.INCLUDING DataSourceTable.name.VALUE"]]
    set columnTypeName [common::getTypeClassName $column]
    
    if [equal $columnTypeName "DBBlob"] {
      set orderColumnsString [concat $orderColumnsString [format "%s%s" "$ds.$columnName" "Path"] $order]
    } {
      set orderColumnsString [concat $orderColumnsString "$ds.$columnName" $order]
    }
    
    if ![equal [llength $orderColumns] $counter] {
      set orderColumnsString [concat $orderColumnsString ","]
    }
    set counter [expr $counter + 1]
  }
  
  return [concat [getHQLQuery $sourceDataSource] "order by" $orderColumnsString]
}

# returns the limit for the given limit-attribute
# attr may be "firstRow" or "maxRows"
proc getLimitExpression {limitDataSource attr} {
  set rowCondition [lindex [c::getList "$limitDataSource.$attr.CHILDREN"] 0]
  set rowString [traverseConditionTree $rowCondition]
  
  if [equal [string index $rowString 0] ":"] {
    return [string replace $rowString 0  0]
  }
  
  return $rowString
}

proc getFirstRowLimit {limitDataSource} {
  return [getLimitExpression $limitDataSource "firstRow"]
}

proc getMaxRowsLimit {limitDataSource} {
  return [getLimitExpression $limitDataSource "maxRows"]
}

proc getLimitHQLQuery {limitDataSource} {
  return [getHQLQuery $limitDataSource]
}

# param type may contain the values "typed" or "stringified"
proc getFormalParameterList {dataSource {type ""}} {
  set className [c::getClassName $dataSource]
  set parameterList ""
  
  if {[equal $className "LimitDataSource"] || [equal $className "DataSourceProjection"]} {
    set parameters [c::getList "$dataSource.formalParameters.CHILDREN"]
    set counter 1
    foreach {param} $parameters {
      set paramName [c::get "$param.name.VALUE"]
      
      if [equal $type "typed"] {
        set parameterList [concat $parameterList [getDataSourceFormalParamWithType $param]]
      } elseif [equal $type "stringified"] {
        set parameterList [concat $parameterList "\"$paramName\""]
      } else {
        set parameterList [concat $parameterList $paramName]
      }
      
      if ![equal [llength $parameters] $counter] {
        set parameterList [concat $parameterList ","]
      }
      set counter [expr $counter + 1]
    }
    
    set parentParams [getFormalParameterList [c::get "$dataSource.sourceDataSource.VALUE"] $type]
    if ![equal $parentParams ""] {
      set parameterList [concat $parameterList "," $parentParams]
    }
  }
  
  if [equal $className "OrderedDataSource"] {
    return [getFormalParameterList [c::get "$dataSource.sourceDataSource.VALUE"] $type]
  }
  
  return $parameterList
}

#######################################################
# NEW NEW NEW
#######################################################
set javaTypeMapping(DBString) "String"
set javaTypeMapping(DBText) "String"
set javaTypeMapping(DBInteger) "int"
set javaTypeMapping(DBLong) "long"
set javaTypeMapping(DBFloat) "float"
set javaTypeMapping(DBDouble) "double"
set javaTypeMapping(DBBoolean) "boolean"
set javaTypeMapping(DBBigInt) "java.math.BigInteger"
set javaTypeMapping(DBDate) "java.util.Date"
set javaTypeMapping(DBTimeStamp) "java.util.Date"
set javaTypeMapping(DBBlob) "org.apache.myfaces.custom.fileupload.UploadedFile"

# returns the type of the primary key of the given "DataSourceTable"
# if no columns are defined "Long" is returned
# if there is a column but no primary key the type of the first column will be returned!
proc getTypeOfPrimaryKey {dsDB} {
  
  variable javaTypeMapping
  set columns [c::getList "$dsDB.columns.CHILDREN"]
  if [equal [llength $columns] 0] {
    return $javaTypeMapping("DBLong")
  }
  
  foreach {column} $columns {
    set primaryKey [c::get "$column.primaryKey.VALUE"]
    if [equal $primaryKey 0] {
      continue
    }
    # we have the primary key
    set typeClass [common::getTypeClassName $column]
    return $javaTypeMapping($typeClass)
  }
  
  set typeClass [common::getTypeClassName [lindex $columns 0]]
  return $javaTypeMapping($typeClass)
}

# returns the column name of the primary key of the given "DataSourceTable"
# if no primarykey is defined the name of the first column is returned
proc getColumnNameOfPrimaryKey {dsDB} {
  
  set columns [c::getList "$dsDB.columns.CHILDREN"]
  if [equal [llength $columns] 0] {
    return "pleaseDefineADBColumnFirst"
  }
  
  foreach {column} $columns {
    set primaryKey [c::get "$column.primaryKey.VALUE"]
    if [equal $primaryKey 0] {
      continue
    }
    # we have the primary key
    return [string tolower [c::get "$column.name.VALUE"]]
  }
  
  # no primary key defined
  # -> return the name of the first column!
  set column [lindex $columns 0]
  return [string tolower [c::get "$column.name.VALUE"]]
}


# returns the HQL-NAMED-WHERE-clause for an databaseModifyOperation
# e.g. "where userid = :userId and lastname = :lastname"
proc getHQLWhereClause {objId} {
  set condition [lindex [c::getList "$objId.conditions.CHILDREN"] 0]
  if ![nullptr condition] {
    return [concat "where " [traverseConditionTree $condition] ]
  }
  return ""
}

proc getInitPropertiesCode {form} {
  
  set dataSource [c::get $form.dataSource.VALUE]
  if [nullptr $dataSource] {
    return ""
  }
  
  set idFormalParam [c::get $form.dataSourceId.VALUE]
  if [nullptr $idFormalParam] {
    return ""
  }
  
  set idName [c::get $idFormalParam.name.VALUE]
  set dsName [string tolower [c::get "$dataSource.name.VALUE"]]
  set DsName [string totitle $dsName]
  
  set initCode "if( FacesUtils.hasRequestParameter(\"$idName\")) {"
  set initCode [concat $initCode "this.$idName =" [getWebNodeFormalRequestParamForJSFBean $idFormalParam] ";\n"]
  set initCode [concat $initCode [format "%s%s%s" "$DsName $dsName = this.getServiceLocator().get" $DsName "Dao().getUserById(this.$idName);"]]
  
  set elements [c::getList $form.elements.CHILDREN]
  foreach {formElement} $elements {
    set initCode [concat $initCode [getInitPropertyMapping $formElement $dsName]]
  }
  
  return [concat $initCode "}" ]
  
}

proc getInitPropertyMapping {formElement objName} {
 set name [string tolower [c::get $formElement.name.VALUE]]
 set dbColumn [c::get $formElement.pvalue.VALUE]
 if [nullptr $dbColumn] {
   return ""
 }
 set Property [string totitle [c::get $dbColumn.name.VALUE]]
 return [format "%s%s%s" "this.$name = $objName.get" $Property "();"]
}

proc getOperationMapping {ds_insert} {
  set ds [string tolower [getTargetDataSourceName $ds_insert]]
  
  set valuesList [c::getList "$ds_insert.values.CHILDREN"]
  set valuesLength [llength $valuesList]
  
  if [equal valuesLength 0] {
    return ""
  }

  set mapping ""
  
  foreach {dataSourceValue} $valuesList {

   set expression [c::getList "$dataSourceValue.expression.CHILDREN"]
   set exprLength [llength $expression]
   if [expr $exprLength <= 0] {
     continue
   }
   
   set column [c::get "$dataSourceValue.column.VALUE"]
   set columnName [string totitle [c::get "$column.name.VALUE"]]
   set columnType [c::getList "$column.type.CHILDREN"]
   set columnTypeClass [c::getClassName $columnType]
   set exprClass [c::getClassName $expression]
   
   if [equal $exprClass "ConstantExpression"] {
       set mapping [concat $mapping [format "%s%s%s%s%s%s" $ds ".set" $columnName "(" [getConstantExpressionValue $expression "\""] ");"]]
       
   } elseif [equal $exprClass "PredefinedExpression"] {
     
     set value [c::get $expression.value.VALUE]
     if [string equal -length 12 $value "current_date"] {
       set mapping [concat $mapping [format "%s%s%s%s" $ds ".set" $columnName "( new Date() );"]]
     } elseif [string equal -length 17 $value "current_timestamp"] {
       # TODO: hier evtl. new Timestamp() ?
       set mapping [concat $mapping [format "%s%s%s%s" $ds ".set" $columnName "( new Date() );"]]
     } {
       set mapping [concat $mapping [format "%s%s%s%s" $ds ".set" $columnName "( $value );"]]
     }
     
       
   } elseif [equal $exprClass "ParameterValueExpression"] {
   
     set formalParameter [c::get "$expression.parameter.VALUE"]
     if ![nullptr $formalParameter] {
        set name [c::get "$formalParameter.name.VALUE"]
        set formalParamType [common::getTypeClassName $formalParameter]
        if [equal $formalParamType "DBBlob"] {
          set mapping [concat $mapping [format "%s%s%s%s%s%s%s" "if( $name != null ) { " $ds ".set" $columnName "ContentType( " $name ".getContentType() );"]]
          set mapping [concat $mapping "String __fileName = PWUtils.saveUploadedFile( this.getUploadedFilesDir(), $name );"]
          set mapping [concat $mapping [format "%s%s%s%s" $ds ".set" $columnName "Path( Constants.UPLOAD_DIR + \"/\" + __fileName ); }"]]
          
        } else {
          set mapping [concat $mapping [format "%s%s%s%s" $ds ".set" $columnName "($name);"]]
        }
     }
   }
 }
 
 return $mapping
 
}

proc getTargetDataSourceName {objId} {
  return [getDataSourceName $objId "targetDataSource"]
}

proc getSourceDataSourceName {objId} {
  return [getDataSourceName $objId "sourceDataSource"]
}

proc getDataSourceName {objId refAttrName} {
  set dataSource [c::get "$objId.$refAttrName.VALUE"]
  if [nullptr $dataSource] {
    return "N/A"  
  }
  return [c::get "$dataSource.name.VALUE"]
}

proc getRefClassName {objId refAttrName} {
  set ref [c::get "$objId.$refAttrName.VALUE"]
  if [nullptr $ref] {
    return "N/A"
  }
  return  [c::getClassName $ref]
}

proc getConstantExpressionValue {objId stringchar} {
   set value [c::get "$objId.value.VALUE"]
   
   if { [string equal $value "null"] || [string equal -length 4 $value "new "] || [string is integer $value] || [string is double $value] || [string is boolean $value] } {
     return $value
   }  
   return [format "%s%s%s" $stringchar $value $stringchar]
}

# Returns the content type of ListOutputRefContent or ListOutputProjectionContent
# Maybe one of "text", "date", "datetime", "image", "misc" at the moment
proc getOutputContentType {outputContent} {
  set listoutputcolumn [c::get "$outputContent.column.VALUE"]
  set listoutputcolumnclass [c::getClassName $listoutputcolumn]
  set dbtype ""
  if [equal $listoutputcolumnclass "DBColumn"] {
    set dbtype [c::getList "$listoutputcolumn.type.CHILDREN"]  
  } {
    # we have a projection
    set dbtype [c::getList "$listoutputcolumn.column.VALUE.type.CHILDREN"]  
  }

  set dbtypeClass [c::getClassName $dbtype]
  
  if [equal $dbtypeClass "DBDate"] {
    return "date"
  }
  
  if [equal $dbtypeClass "DBTimeStamp"] {
    return "datetime"
  }
  
  if ![equal $dbtypeClass "DBBlob"] {
    return "text"
  }
  
  # we have a blob column -> return assoc. content type
  return [c::get "$dbtype.contentType.VALUE"]
}

proc getDataSourceFormalParamWithType {dataSourceFormalParam} {
  variable javaTypeMapping
  
  set name [c::get "$dataSourceFormalParam.name.VALUE"]
  set typeClass [common::getTypeClassName $dataSourceFormalParam]  
  set javatype $javaTypeMapping($typeClass)
  
  return [concat $javatype $name]
}

proc getWebNodeFormalParamWithType {formalParam} {
  variable javaTypeMapping
  set name [c::get "$formalParam.name.VALUE"]
  set type [c::getList "$formalParam.type.CHILDREN"]
  set typeClass [c::getClassName $type]  
  set javatype $javaTypeMapping($typeClass)
  
  return [concat $javatype $name ]
}

proc getJavaTypeOfWebNodeFormalParam {formalParam} {
  variable javaTypeMapping
  set type [c::getList "$formalParam.type.CHILDREN"]
  set typeClass [c::getClassName $type]  
  return $javaTypeMapping($typeClass)
}

proc getWebNodeActualParamWithType {webNodeActualParam} {
  set formalParam [c::get "$webNodeActualParam.formalParam.VALUE"]
  return [getWebNodeFormalParamWithType $formalParam]
}

proc getTypedRequestParam {dbtype retVal} {
    set type [c::getList "$dbtype.CHILDREN"]
    set typeClass [c::getClassName $type]
    
    if {[equal $typeClass "DBString" ] || [equal $typeClass "DBText" ]} {
      return $retVal;
    } elseif [equal $typeClass "DBInteger" ] {
      return [concat "Integer.parseInt(" $retVal ")"]
    } elseif [equal $typeClass "DBLong" ] {
      return [concat "Long.parseLong(" $retVal ")"]
    } elseif [equal $typeClass "DBFloat" ] {
      return [concat "Float.parseFloat(" $retVal ")"]
    } elseif [equal $typeClass "DBDouble" ] {
      return [concat "Double.parseDouble(" $retVal ")"]
    } elseif [equal $typeClass "DBBoolean" ] {
      return [concat "Boolean.parseBoolean(" $retVal ")"]
    }
    
    #puts "WARN:$typeClass not parseable yet. setting to null!"
    return "null"
}

proc getWebNodeSessionExpression {sessExpr} {
  set dbColumn [c::get $sessExpr.content.VALUE]
  if [nullptr $dbColumn] {
    return "null"
  }
  
  set Property [string totitle [c::get $dbColumn.name.VALUE]]
  set Ds [string totitle [c::get "$dbColumn.INCLUDING DataSourceTable.name.VALUE"]]
  
  if [equal [common::getTypeClassName $dbColumn] "DBBlob"] {
    return [format "%s%s%s" "(($Ds)FacesUtils.getSessionAttribute(Key.USER)).get" $Property "Path()"]
  }
  return [format "%s%s%s" "(($Ds)FacesUtils.getSessionAttribute(Key.USER)).get" $Property "()"]
}

proc getWebNodeSessionExpressionForServlet {sessExpr} {
  set dbColumn [c::get $sessExpr.content.VALUE]
  if [nullptr $dbColumn] {
    return "null"
  }
  
  set Property [string totitle [c::get $dbColumn.name.VALUE]]
  set Ds [string totitle [c::get "$dbColumn.INCLUDING DataSourceTable.name.VALUE"]]

  if [equal [common::getTypeClassName $dbColumn] "DBBlob"] {
    return [format "%s%s%s" "(($Ds)request.getSession().getAttribute(Key.USER)).get" $Property "Path()"]
  } 
  return [format "%s%s%s" "(($Ds)request.getSession().getAttribute(Key.USER)).get" $Property "()"]
}

proc getWebNodeVariableExpression {webNodeVariableExpression} {
  
  set varDef [c::get "$webNodeVariableExpression.content.VALUE"]
  set varDefClass [c::getClassName $varDef]
  set varDefName [c::get "$varDef.name.VALUE"]
  set retVal "request.getParameter(\"$varDefName\" )"
  
  if [equal $varDefClass "WebNodeFormalParam"] {
    return [getTypedRequestParam "$varDef.type" $retVal]
  }
  # TODO evtl fr FormularElement auch noch
  # dabei anhand des assoc. converters feststellen, um was fr einen Typ es sich handelt
  
  return $retVal
}

proc getParamExpressionValue {listOutputActualParam} {
  set expression [c::getList $listOutputActualParam.expression.CHILDREN]
  set exprClassName [c::getClassName $expression]
  if {[equal $exprClassName "ColumnValueExpression"] || [equal $exprClassName "ProjectionColumnValueExpression"]} {
    return [string tolower [c::get $expression.column.VALUE.name.VALUE]]
  }
  return "Error: $exprClassName not allowed here!"
}


proc getListOutputActualRequestParam {loActualParam} {
  return [getWebNodeFormalRequestParam [c::get "$loActualParam.formalParam.VALUE"]]
}

proc getWebNodeFormalRequestParam {formalParam} {
  set formalParamName [c::get "$formalParam.name.VALUE"]
  set retVal "request.getParameter(\"$formalParamName\" )"
  return [getTypedRequestParam "$formalParam.type" $retVal]
}

proc getWebNodeFormalRequestParamForJSFBean {formalParam} {
  set formalParamName [c::get "$formalParam.name.VALUE"]
  set retVal "FacesUtils.getRequestParameter(\"$formalParamName\" )"
  return [getTypedRequestParam "$formalParam.type" $retVal]
}


proc isUsedInWebPage {formalParam} {
  #if {[isUsedByWebAction $formalParam] || [isUsedInOutputFragment $formalParam]} {
  #  return true
  #}
  
  return true
}

proc isUsedInOutputFragment {formalParam} {
  set webpage [c::get "$formalParam.PARENT.PARENT"]
  if ![equal [c::getClassName $webpage] "WebPage"] {
    return false
  }
  
  set components [c::getList "$webpage.frame.CHILDREN.components.CHILDREN"]
  foreach {comp} $components {
    if ![equal [c::getClassName $comp] "Paragraph"] {
      continue
    }
    
    set fragments [c::getList $comp.fragments.CHILDREN]
    foreach {fragment} $fragments {
      if ![equal [c::getClassName $fragment] "OutputFragment"] {
        continue
      }
      
      if [equal $formalParam [c::get $fragment.content.VALUE]] {
        return true
      }
    }
  }
  
  return false
}

proc isUsedByWebAction {formalParam} {
  set webpage [c::get "$formalParam.PARENT.PARENT"]
  if ![equal [c::getClassName $webpage] "WebPage"] {
    return false
  }
  
  set components [c::getList "$webpage.frame.CHILDREN.components.CHILDREN"]
  foreach {comp} $components {
    if ![equal [c::getClassName $comp] "Formular"] {
      continue
    }
    
    set actualParams [c::getList "$comp.actualParams.CHILDREN"]
    foreach {actualParam} $actualParams {
      set expression [lindex [c::getList "$actualParam.expressions.CHILDREN"] 0]
      if ![equal [c::getClassName $expression] "WebNodeVariableExpression"] {
        continue
      }
      set content [c::get $expression.content.VALUE]
      if ![equal [c::getClassName $content] "WebNodeFormalParam"] {
        continue
      }
      if [equal $formalParam $content] {
        return true
      }
    }
  }
  return false
}

proc getAttribute {obj} {

    set valueObject [c::getList "$obj.value.CHILDREN"]
    set category [c::get $obj.category]

    if {[equal category Padding]} {
        # size
        return "padding-"
    }
}


proc getErrorLink {objId} {
        set className [c::getClassName $objId]

        if {[equal $className "WebPage"] || [equal $className "WebAction"]} {
                set webzone_obj  [c::get $objId.PARENT.PARENT]
                set error_link_obj [c::get $webzone_obj.errorLink.VALUE]
                if [nullptr $error_link_obj] {
                  return ""
                }
                
                set error_link_name [c::get $error_link_obj.name.VALUE]
                set error_link_class [c::getClassName $error_link_obj]
                if [equal $error_link_class "WebPage"] {
                  return [format "%s%s" $error_link_name ".jsf"]
                }
                
                return $error_link_name
        }

        return ""
}


proc getRoles {objId} {
    set roleList ""
    set roles [c::getList "$objId.roles.CHILDREN"]
    foreach {listEntry} $roles {
       set roleList [concat $roleList [c::get $listEntry.role.VALUE.name.VALUE]]
    }
    return $roleList
}

proc getWebpageName {objId} {
 set webpage [c::get "$objId.INCLUDING WebPage.name.VALUE"]
 if ![nullptr $webpage] {
  return $webpage
 } {
  return "undefined"
 }
}

# computes the role list for an action connection
proc getRoleListOfActionConnection {objId} {
  return [getRoleList [c::get $objId.action.VALUE]]
}

proc hasRoleRestriction {objId} {
  set roles [getRoleListOfActionConnection $objId]
  if [equal [llength $roles] 0] { return false }
  return true
}

# Computes the role list for a web node
proc getRoleList {objId} {
  set className [c::getClassName $objId]
  
  if {[equal $className "WebPage"] || [equal $className "WebAction"]} {
    set webzone_obj  [c::get $objId.PARENT.PARENT]
    # special case: free zones gets an complete empty permission_list
    set roles [getRoles $webzone_obj]
    if [equal $roles ""] {
      # free zone
      return ""
    }
    return [getRoleList $webzone_obj]
  }
  
  if {[equal $className "SubWebZone"] || [equal $className "RootWebZone"]} {
    set roleList [getRoles $objId]
    
    # go down
    set target_list [c::getList $objId.webTargets.CHILDREN]
    foreach {listEntry} $target_list {
      set name [c::getClassName $listEntry]
        if {[equal $name "SubWebZone"] || [equal $name "RootWebZone"]} {
          set roleList [concat $roleList [getRoleList $listEntry]]
        }
    }
    return $roleList
  }
  
  return ""
}

proc getStyleSheetBlock {objId} {
  return [getStyleSheetClass $objId "styleSheetBlock"]
}

proc getStyleSheetClass {objId refname} {
  set StyleSheetBlock [c::getOrNull "$objId.$refname.VALUE"]
  if ![nullptr $StyleSheetBlock] {
    set StyleSheetBlock [c::getOrNull "$StyleSheetBlock.name.VALUE"]
    return $StyleSheetBlock
  } else {
    return ""
  }
}

# evtl hilfreich: [file join $::main::appHome "fixCode"]
proc copyFile {src_file dst_dir} {
  
  if {![string equal -length 1 $src_file "/"] && ![equal [string match "*:*" $src_file] 1] } {
    set src_file [file join "examples" $src_file]
  }
  
  #log::log "XXX" [concat "copying" $src_file "to" $dst_dir]
  
  if [file exists $src_file] {
    if [file exists [file join $dst_dir [file tail $src_file]] ] {
      return
    }
    
    file copy -force $src_file $dst_dir
  }
}

proc toPath {pkg_name} {

    set splitted [split $pkg_name .] 
    set pkg_path [join $splitted /]
    return $pkg_path
    
}

proc execAnt {outputdir target} {
    exec ant -f [format "%s%s" $outputdir "/build.xml"] $target > [format "%s%s" $outputdir "/ant.log"]
}
