
# Modul: activeCanvas
#
# Requires: vec, misc.
#
# activeCancas ist ein "high level" Canvas inklusive Scrollbalken. Die Abstraktionsebene im Vergeleich
# zum normalen Canvas ist wesentlich hher. Logischen Objekten ("Gruppen") knnen aus mehreren Primitiven zusammengesetzt
# werden. Gruppen knnen interaktive Eigenschaften zugeordnet werden.
#
# activeCanvas untersttz folgende Konpzepte:
#
#  - Gruppen = Mengen von Canvas-Primitiven
#  - Selektion von Gruppen
#  - Zuordnung von Tags zu Gruppen
#  - Drag-and-Drop (eine Gruppe ist dabei das Drag-Objekt, eine andere das Drop-Objekt)
#  - Rotation von Drop-Selektionen (falls mehrere visuell bereinander liegen)
#  - Aktion durch Doppelklick
#  - Kontextmen ffnen durch Rechtsklick
#  - Drag-and-Drop ber Fenstergrenzen hinweg
#
# Es fehlt noch: Mehrfachselektion!
#
# Gruppen besitzen folgende Eigenschaften:
#
#  - Haben eine beliebige Menge von Attribut/Wert - Tags
#  - wissen, auf welche Objekte sie "gedropped" werden knnen und welche Wirkung das hat
#  - Haben einen der folgenden Darstellungszustnde: Normal, Selected, Highlighted
# 
# Canvas-Primitive besitzen folgende Eigenschaften:
#
#  - clickable
#  - dropable
#  - Darstellungszustand
#
# Das Modul ist unabhngig von der Implementierung der zugrundeliegenden abstrakten Struktur, d.h. auch
# Auerhalb von DEViL anwendbar. Gltigkeit von von Drop-Operationen, Eintrge des Kontextmens, sowie 
# Auswirkungen von Operationen knnen Mittels Callback-Funktionen beschrieben werden.


namespace eval ::activeCanvas {
}


#######################################################################################  Koordinatenumrechnung

# Verschiedene Arten der Koordinaten
#
# * screen: Absolute Bildschirmkoordinaten
# * window (win): Ursprung liegt links oben im Canvas-Viewport
# * primitive (prim): Koordinatensystem der Canvas-Primitive (ndert sich bei verschiedenen Zoom-Faktoren)
# * logical (log): Logische Koordinaten der Canvas-Primitve (unabhngig von Zoom-Faktor)
#
# Philosophie:
# * Falls mglich wird mit logischen Koordinaten hantiert!


proc activeCanvas::getZoom {obj} {
    upvar 0 ::activeCanvas::persData_$obj persData
    return $persData(zoom)
}


proc activeCanvas::setZoom {obj zoom} {
    upvar 0 ::activeCanvas::persData_$obj persData
    set persData(zoom) $zoom
}


proc activeCanvas::zoomIn {obj} {
    setZoom $obj [expr {[getZoom $obj]*2.0}]
}


proc activeCanvas::zoomOut {obj} {
    setZoom $obj [expr {[getZoom $obj]/2.0}]
}


proc activeCanvas::screenCoords2winCoords {obj absPos} {
    upvar 0 ::activeCanvas::$obj data
    assign "x y" $absPos
    set dx [winfo rootx $data(canvas)]
    set dy [winfo rooty $data(canvas)]
    return "[expr {$x - $dx}] [expr {$y - $dy}]" 
}


proc activeCanvas::winCoord2primCoord {obj winCoord} {
    upvar 0 ::activeCanvas::$obj data
    set primCoord ""
    foreach {x y} $winCoord {
        lappend primCoord [$data(canvas) canvasx $x] [$data(canvas) canvasy $y]
    }
    return $primCoord
}


proc activeCanvas::primCoord2logCoord {obj primCoord} {
    set zoom [getZoom $obj]
    set logCoord ""
    foreach {x y} $primCoord {
        lappend logCoord [expr { round($x/$zoom) }] [expr { round($y/$zoom) }]
    }
    return $logCoord
}


proc activeCanvas::winCoord2logCoord {obj winCoord} {
    return [primCoord2logCoord $obj [winCoord2primCoord $obj $winCoord]]
}


proc activeCanvas::logCoord2primCoord {obj logCoord} {
    set zoom [getZoom $obj]
    set primCoord ""
    foreach {x y} $logCoord {
        lappend primCoord [expr { round($x*$zoom) }] [expr { round($y*$zoom) }]
    }
    return $primCoord
}


########################################################### Operationen mit nicht-logischen Koordinaten



proc activeCanvas::selectPointedElement {obj windowPos {multi 0}} {
    #sorgt dafuer, dass das angeklickte Element gewaehlt wird
    set logPos [winCoord2logCoord $obj $windowPos]
    setSelection $obj [getPointedElement $obj $logPos] $multi
}


proc activeCanvas::openContextMenu {obj windowPos X Y} {
    upvar 0 ::activeCanvas::$obj data

    #on_buttonPress1 $obj $windowPos
    selectPointedElement $obj $windowPos
    set group [getSelection $obj]
    if { ![empty $group] } {
        eval $data(onContextMenu,$group) $group $X $Y
    }
}


proc activeCanvas::on_buttonPress3 {obj windowPos X Y} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        objectSelection {
            openContextMenu $obj $windowPos $X $Y
        }
        default {
            returnToObjectSelection $obj
        }
    }
}


proc activeCanvas::on_doubleClick {obj windowPos} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        objectSelection {
            startDoubleClickAction $obj
        }
        default {
            returnToObjectSelection $obj
        }
    }
}

proc activeCanvas::on_buttonPressOutside {obj} {
    upvar 0 ::activeCanvas::$obj data
    if {$data(pressedInside) == 0} {
        eval $data(toolbarCallback)
        activeCanvas::returnToObjectSelection $obj
    }
    set data(pressedInside) 0
}

proc activeCanvas::on_buttonPressAndControllKey {obj windowPos} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        objectSelection {
            selectPointedElement $obj $windowPos 1
        }
        default {
            returnToObjectSelection $obj
        }
    }
}


proc activeCanvas::on_buttonPress1 {obj windowPos} {

    upvar 0 ::activeCanvas::$obj data
    switch -regexp $data(interactionState) {
        objectSelection {
            selectPointedElement $obj $windowPos
            if { ![empty $data(clickedGroup)] } {
                setIaState $obj maybeDraging
                set data(interWindowDragDest) $obj
            }
            set group [getSelection $obj]
            if { ![empty $group] } {
                if {[info exists data(attr-clickAttr,clickAble-$group)]} {
                     set clickAttr [getAttr $obj clickAble-$group clickAttr]
                     eval $data(onClick,clickAble-$group) $clickAttr
                }
            }
        }
        connecting {
            set data(firstConnectionPoint) $data(dropgroup)
            set data(firstConnectionPos) [getGroupPosition $obj $data(dropgroup)]
            setIaState $obj connecting2
            set data(acceptDropPredicate) $data(acceptDropPredicate2)
            startShowingLinePreview $obj $data(firstConnectionPos) $data(dropPos) 
        }
        inspointSelection {
            selectDropPoint $obj
            eval $data(toolbarCallback)
            returnToObjectSelection $obj
        }
        default {
            log::error "on_buttonPress1: Invalid state detected: $data(interactionState)"
            returnToObjectSelection $obj
        }
    }
    set data(pressedInside) 1
}


proc activeCanvas::on_b1Motion {obj windowPos absPos} {
    upvar 0 ::activeCanvas::$obj data
    
    set pos [winCoord2logCoord $obj $windowPos]
    
    switch $data(interactionState) {
        objectSelection {
            return
        }
        inspointSelection {
            updateDropPointSet $obj $pos
        }
        draging {
            set destCanvas [eval winfo containing $absPos]
            if { [info exists activeCanvas::activeCanvasId($destCanvas)] } {
                set destCanvas $activeCanvas::activeCanvasId($destCanvas)
            } else {
                set destCanvas $obj
            }
            if { ![equal $destCanvas $data(interWindowDragDest)] } { 
                updateDropPointSet $data(interWindowDragDest) ""
                set data(interWindowDragDest) $destCanvas
                if { ![equal $destCanvas $data(canvas)] } {
                    startAbstractDrop $destCanvas $data(acceptDropPredicate) $data(onDrop) $data(clickedGroup)
                }
            }
            
            updateDropPointSet $destCanvas [winCoord2logCoord $obj [screenCoords2winCoords $obj $absPos]]
        }
        connecting {
            updateDropPointSet $obj $pos
        }
        connecting2 {
            updateDropPointSet $obj $pos
            showLinePreview $obj $data(firstConnectionPos) $pos
        }
        maybeDraging {
            if { [vec::distance $data(clickpos) $pos] > $cfg::moveTreshold } {
                setIaState $obj draging
                updateDropPointSet $obj $pos
            }
        }
        default {
            log::error "activeCanvas::on_b1Motion: Invalid state detected"
            returnToObjectSelection $obj
        }
    }
}



###################################################################################### basic operations


proc activeCanvas::removeOptions {toRemove inputList} {
    set result ""
    foreach {attr val} $inputList {
        if { ![elementOf $attr $toRemove] } {
            lappend result $attr $val
        }
    }
    return $result
}


proc activeCanvas::removeUnchangedOptions {optionList changeOptionList} {
    foreach {opt value} $changeOptionList {
        set change($opt) 1
    }
    set result ""
    foreach {opt value} $optionList {
        if { [info exists change($opt)] } {
            lappend result $opt $value
        }
    }
    return $result
}


proc activeCanvas::getZoomedFont {obj origFont} {
    upvar 0 ::activeCanvas::persData_$obj persData
    set zoom [getZoom $obj]
    if {![info exists persData(zoomedFont,$origFont,$zoom)]} {
        array set options [font actual $origFont]
        set options(-size) [expr { round( [getZoom $obj] * $options(-size)) }]
        set newFont [eval font create [array get options]]
        set persData(zoomedFont,$origFont,$zoom) $newFont
        return $newFont
    } else {
        return $persData(zoomedFont,$origFont,$zoom)
    }
}


proc activeCanvas::getZoomedImage {obj origImage} {
    upvar 0 ::activeCanvas::persData_$obj persData
    set zoom [getZoom $obj]
    if {![info exists persData(zoomedImage,$origImage,$zoom)]} {
        set newImage [image create photo]
        if {$zoom>1} {
            $newImage copy $origImage -zoom [expr {round([getZoom $obj])}]
        } else {
            $newImage copy $origImage -subsample [expr {round(1.0/[getZoom $obj])}]
        }
        set persData(zoomedImage,$origImage,$zoom) $newImage
        return $newImage
    } else {
        return $persData(zoomedImage,$origImage,$zoom)
    }
}

proc activeCanvas::measureWindowObject {obj type params} {
    upvar 0 ::activeCanvas::$obj data
    eval $type $data(canvas).windPrimitiveTESTOBJ $params
    set height [winfo reqheight $data(canvas).windPrimitiveTESTOBJ]
    set width [winfo reqwidth $data(canvas).windPrimitiveTESTOBJ]
    destroy $data(canvas).windPrimitiveTESTOBJ
    return "$width $height"
}

proc activeCanvas::createWindowObject {obj type pos params} {
    upvar 0 ::activeCanvas::$obj data
    set pos [logCoord2primCoord $obj $pos]
    incr data(countWindPrim)
    lappend data(winObjectGrp) $data(canvas).windPrimitive$data(countWindPrim)
    if {[catch {eval $type $data(canvas).windPrimitive$data(countWindPrim) $params} errMessage ]} {
        tk_messageBox -message $errMessage
    }
    set result [eval $data(canvas) create window $pos -window $data(canvas).windPrimitive$data(countWindPrim)]
    return $result
}


proc activeCanvas::createPrimitive {obj type pos params} {
    upvar 0 ::activeCanvas::$obj data
    set result ""
    set pos [logCoord2primCoord $obj $pos]
    if { [procExists ::canvasitems::$type] } {
      
        foreach {iType iPos iParams} [::canvasitems::$type $pos $params] {
           
            lappend result [eval $data(canvas) create $iType $pos $iParams]
           
        }        
    } else {
        if { [equal $type "image"] && [getZoom $obj]!=1 } {
            array set param $params
            set param(-image) [getZoomedImage $obj $param(-image)]
            set params [array get param]
        }
        if { [equal $type "text"] && [getZoom $obj]!=1 } {
            array set param $params
            set param(-font) [getZoomedFont $obj $param(-font)]
            set params [array get param]
        }
        
        lappend result [eval $data(canvas) create $type $pos $params]
        
    }
    return $result
}


xn::proc activeCanvas::createObject {obj group type pos -normal "" -select "" -highlight "" -tags "" -level "0" -rotate 0} {
    upvar 0 ::activeCanvas::$obj data
    
    >> xn::assert { $pos ne "" }
    lappend tags "level-$level"
    foreach cObj [createPrimitive $obj $type $pos [concat [removeOptions "-clickable -dropable" $normal] -tags [list $tags]]] {
        set data(obj-$cObj,normalAttrs) [removeUnchangedOptions $normal [concat $select $highlight "-clickable xxx -dropable xxx"]]
        set data(obj-$cObj,selectAttrs) $select
        set data(obj-$cObj,highlightAttrs) $highlight
        setTagValue $obj $cObj group $group
        setTagValue $obj $cObj rotate $rotate
        setElemState $obj $cObj normal
    }
    return $cObj
}


proc activeCanvas::setAttributes {obj cObj list} {
    upvar 0 ::activeCanvas::$obj data

    foreach {attr val} $list {
        switch -- $attr {
            "-clickable" {
                setTagValue $obj $cObj clickable $val
            }
            "-dropable" {
                setTagValue $obj $cObj drop $val
            }
            default {
                $data(canvas) itemconfigure $cObj $attr $val
            }
        }
    }
}

    
xn::proc activeCanvas::setElemState {obj tag state} {
    upvar 0 ::activeCanvas::$obj data
    >> xn::assert { [elementOf $state "normal select highlight"] }
    foreach cObj [$data(canvas) find withtag $tag] {
        setAttributes $obj $cObj $data(obj-$cObj,normalAttrs)
        setAttributes $obj $cObj $data(obj-$cObj,${state}Attrs)
    }
}


proc activeCanvas::setGroupState {obj group state} {
    upvar 0 ::activeCanvas::$obj data
    upvar 0 ::activeCanvas::$obj-states states
    
    set states($group) $state
    setElemState $obj group-$group $state
}


proc activeCanvas::getGroupsWithState {obj state} {
    upvar 0 ::activeCanvas::$obj data
    upvar 0 ::activeCanvas::$obj-states states

    set result ""
    foreach {group gstate} [array get states] {
        if { [equal $gstate $state] } {
            lappend result $group
        }
    }
    return $result
}


xn::proc activeCanvas::getTagValue {obj cObj tagname} {
    upvar 0 ::activeCanvas::$obj data

    >> xn::assert { $cObj ne "" }
    set tags [$data(canvas) itemcget $cObj -tags]
    set index [lsearch $tags "$tagname-*"]
    >> xn::assert { $index != -1 }
    set tag [lindex $tags $index]
    set value [join [lrange [split $tag "-"] 1 end] "-"]
    return $value
}


proc activeCanvas::setTagValue {obj cObj tagname value} {
    upvar 0 ::activeCanvas::$obj data

    set tags [$data(canvas) itemcget $cObj -tags]
    set index [lsearch $tags "$tagname-*"]
    set newtags [lreplace $tags $index $index $tagname-$value]
    $data(canvas) itemconfigure $cObj -tags $newtags
}


proc activeCanvas::hasTag {obj cObj tag} {
    upvar 0 ::activeCanvas::$obj data

    set tags [$data(canvas) itemcget $cObj -tags]
    return [elementOf $tag $tags]
}


proc activeCanvas::getDropTargetGroups {self} {
    # benoetigt, um Aktivierung von der Toolbar-Knoepfe effizienter realisieren
    # zu koennen als mittels "canInsert".

    upvar 0 ::activeCanvas::$self data
    set result ""
    foreach cObj [$data(canvas) find withtag drop-1] {
        lappend result [getTagValue $self $cObj group]
    }
    return $result
}


proc activeCanvas::findObjects {self tag predicate} {
    upvar 0 ::activeCanvas::$self data

    set result ""
    foreach cObj [$data(canvas) find withtag $tag] {
        set group [getTagValue $self $cObj group]
        if { [eval $predicate $group] } {
            lappend result $cObj
        }
    }
    return $result
}


proc activeCanvas::findNearestSet {obj pos tag predicate} {
    upvar 0 ::activeCanvas::$obj data

    set nearestDist "1000000"
    set nearestList ""

    foreach cObj [findObjects $obj $tag $predicate] {
        set group [getTagValue $obj $cObj group]
        set dist [vec::distancePointRect $pos [primCoord2logCoord $obj [$data(canvas) bbox $cObj]]]
        if { $dist < $nearestDist } {
            set nearestDist $dist
            set nearestList ""
        }
        if { $dist == $nearestDist } {
            lappend nearestList $cObj
        }
    }
    if { ![empty $nearestList] && [getTagValue $obj [lindex $nearestList end] rotate]==0 } {
        return [lindex $nearestList end]
    } else {
        return $nearestList
    }
}


proc activeCanvas::findNearest {obj pos tag predicate} {
    set nearestList [findNearestSet $obj $pos $tag $predicate]
    return [lindex $nearestList end]
}


proc activeCanvas::findObjectAt {obj pos tag} {
    upvar 0 ::activeCanvas::$obj data
    set primPos [logCoord2primCoord $obj $pos]
    set cObjects [lreverse [eval $data(canvas) find overlapping [vec::makeSquare $primPos $cfg::clickBox]]]
    foreach cObj $cObjects {
        if { [hasTag $obj $cObj $tag] } {
            return $cObj
        }
    }
    return ""
}


###################################################################################### abstract interaction


proc activeCanvas::startDoubleClickAction {obj} {
    upvar 0 ::activeCanvas::$obj data

    set group $data(clickedGroup)
    if { ![empty $group] } {
        if { ![equal $data(onDoubleClick,$group) "activeCanvas::onDoubleClick_nothing"] } {
            eval $data(onDoubleClick,$group) $group
            return
        }
    }
    set group [getSelection $obj]
    if { ![empty $group] } {
        eval $data(onDoubleClick,$group) $group
    }
}

proc activeCanvas::getPointedElement {obj pos} {
    upvar 0 ::activeCanvas::$obj data
    set data(clickpos) $pos
    set data(dropPos) $pos
    set data(drag) 0
    set data(dropgroup) ""

    set data(clickedCObj) [findObjectAt $obj $pos clickable-1]
    if { ![empty $data(clickedCObj)] } {
        set data(clickedGroup) [getTagValue $obj $data(clickedCObj) group]
        set data(acceptDropPredicate) "$data(acceptDropFkt,$data(clickedGroup)) $obj $data(clickedGroup)"
        set data(onDrop) $data(onDrop,$data(clickedGroup))
    } else {
        set data(clickedGroup) ""
    }
    return $data(clickedGroup)
}



proc activeCanvas::updateDropPointSet {obj pos} {
    upvar 0 ::activeCanvas::$obj data
    if { ![empty $pos] } {
        set nearestSet [findNearestSet $obj $pos drop-1 $data(acceptDropPredicate)]
    } else {
        set nearestSet ""
    }
    set data(dropPointMousePos) $pos
    if { ![equal $nearestSet $data(selectedDropPointSet)] } {
        set data(selectedDropPointSet) $nearestSet
        set data(dropPointIndex) 0
        highlightDropPoint $obj
        rotateDropPoint $obj
    } else {
        highlightDropPoint $obj
    }
}


proc activeCanvas::rotateDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data
    if { ![empty $data(dropPointRotationTimer)] } {
        after cancel $data(dropPointRotationTimer)
    }
    if { ![empty $data(selectedDropPointSet)] } {
        set len [llength $data(selectedDropPointSet)]
        set data(dropPointIndex) [expr ($data(dropPointIndex)+$len+-1) % $len]
        set data(dropPointRotationTimer) [after $::cfg::rotationDelay [list activeCanvas::rotateDropPoint $obj]]
        highlightDropPoint $obj
    }        
}    

proc activeCanvas::unhighlightDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data
    foreach group [getGroupsWithState $obj highlight] {
        setGroupState $obj $group normal
    }
    $data(canvas) delete group-temp
    set data(dropgroup) ""

}

proc activeCanvas::highlightDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data
    unhighlightDropPoint $obj
    if { [empty $data(selectedDropPointSet)] } {
        return
    }
    set pos $data(dropPointMousePos)
    set nearest [lindex $data(selectedDropPointSet) $data(dropPointIndex)]

    set group [getTagValue $obj $nearest group]
    setGroupState $obj $group highlight
    set data(dropgroup) $group
    set oldpos $data(dropPos)
    set data(dropPos) $pos
    set data(dropDelta) [vec::sub $pos $data(clickpos)]
    eval $data(onVisualDrop,$group) [list $obj $group $data(clickpos) $pos [vec::sub $pos $oldpos]]
}


proc activeCanvas::selectDropPoint {obj} {
    upvar 0 ::activeCanvas::$obj data
   
    if { ![empty $data(dropgroup)] } {
        eval $data(onDrop) [list $obj $data(clickedGroup) $data(dropgroup) $data(dropPos) $data(dropDelta)]
    }
    updateDropPointSet $obj ""
}


# XXX
proc activeCanvas::startAbstractDrop {obj acceptDropPredicate onDrop clickedGroup} {
    upvar 0 ::activeCanvas::$obj data
    set data(acceptDropPredicate) $acceptDropPredicate
    set data(dropPos) "0 0"
    set data(clickpos) "0 0"
    set data(clickedGroup) $clickedGroup
    set data(onDrop) $onDrop

    # 2. updateDropPointSet
    # 3. selectDropPoint
}



###################################################################################### concrete interaction


# Werte von $data(interactionState)   |  Verfuegbare Variablen
# ---------------------------------------------------------------------------
#   objectSelection
#   maybeDraging                         clickpos, clickedCObj, clickedGroup
#   draging                              clickpos, clickedCObj, clickedGroup, dropPos, dropgroup
#   inspointSelection                    clickpos, clickedCObj, clickedGroup, dropPos, dropgroup


proc activeCanvas::setIaState {obj state} {
    upvar 0 ::activeCanvas::$obj data    
    set data(interactionState) $state
}


proc activeCanvas::canInsert {self acceptDropPredicate} {
    upvar 0 ::activeCanvas::$self data
    set found [findObjects $self drop-1 $acceptDropPredicate]
    return [expr [llength $found]>0]
}


proc activeCanvas::startObjectInsertion {obj acceptDropPredicate onDrop  {toolbarCallback ""}} {
    upvar 0 ::activeCanvas::$obj data
    returnToObjectSelection $obj
    setIaState $obj inspointSelection
    set data(acceptDropPredicate) $acceptDropPredicate
    set data(dropPos) "0 0"
    set data(clickpos) "0 0"
    set data(clickedGroup) "NOGROUP"
    set data(onDrop) $onDrop
    set data(toolbarCallback) $toolbarCallback
}

proc activeCanvas::startConnectionInsertion {obj acceptDropPredicate1 acceptDropPredicate2 onConnect {toolbarCallback ""}} {   
    upvar 0 ::activeCanvas::$obj data
    returnToObjectSelection $obj
    setIaState $obj connecting    
    set data(acceptDropPredicate) $acceptDropPredicate1
    set data(acceptDropPredicate2) $acceptDropPredicate2
    set data(dropPos) "0 0"
    set data(clickpos) "0 0"
    set data(clickedGroup) "NOGROUP"
    set data(onDrop) ""
    set data(onConnect) $onConnect
    set data(toolbarCallback) $toolbarCallback
}



proc activeCanvas::returnToObjectSelection {obj} {
    upvar 0 ::activeCanvas::$obj data
    updateDropPointSet $obj ""
    setIaState $obj objectSelection
}


proc activeCanvas::on_buttonRelease {obj} {
    upvar 0 ::activeCanvas::$obj data
    switch $data(interactionState) {
        maybeDraging {
            setIaState $obj objectSelection
        }     
        connecting2 {
            endShowingLinePreview $obj
            eval $data(onConnect) $obj $data(firstConnectionPoint) $data(dropgroup)
            returnToObjectSelection $obj
            eval $data(toolbarCallback)
        }
        draging {
            selectDropPoint $data(interWindowDragDest)
            setIaState $obj objectSelection

        }
    }
}


############################################################################### some default implementations


proc activeCanvas::onVisualDrop_insertionCursor {cursorSize obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data
    set cursor [primCoord2logCoord $obj [$data(canvas) bbox group-$data(clickedGroup)]]
    if { [empty $cursor] } {
        set cursor "$pos [vec::add $pos $cursorSize]"
    } else {
        set cursor [vec::move $cursor [vec::sub $pos $startPos]]
    }
    set cursor [vec::moveInside $cursor [primCoord2logCoord $obj [$data(canvas) bbox group-$group]]]

    createObject $obj temp rectangle $cursor  -normal "-width 2"
    set data(dropPos) [lrange $cursor 0 1]
}


proc activeCanvas::onVisualDrop_moveArrow {obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data
    assign "x y dummy dummy" [vec::moveInside "$pos $pos" [primCoord2logCoord $obj [$data(canvas) bbox group-$group]]]
    createObject $obj temp line "$startPos $x $y" -normal "-arrow last"
}


proc activeCanvas::onVisualDrop_moveArrowFullScreen {obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data

    createObject $obj temp line "$startPos $pos" -normal "-arrow last"
}


proc activeCanvas::onVisualDrop_createAssoc {obj group startPos pos delta} {
    upvar 0 ::activeCanvas::$obj data

    createObject $obj temp line "$startPos $pos" -normal "-dash {4 4}"
}

proc activeCanvas::onVisualDrop_movePreview {obj group startPos pos delta} {
    # attribute "visualDragParams" of the draged group must be defined!
    upvar 0 ::activeCanvas::$obj data
    assign "dx dy" $delta
    foreach cObj [$data(canvas) find withtag group-$data(clickedGroup)] {
        set newCoords ""
        foreach {x y} [primCoord2logCoord $obj [$data(canvas) coords $cObj]] {
            lappend newCoords [expr $x+$dx] [expr $y+$dy]
        }
        $data(canvas) coords $cObj [logCoord2primCoord $obj $newCoords]
    }
    foreach {cObj pointIndex} [getAttr $obj $data(clickedGroup) "visualDragParams"] {
        set coords [primCoord2logCoord $obj [$data(canvas) coords $cObj]]
        set x [lindex $coords [expr 2*$pointIndex]]
        set y [lindex $coords [expr 2*$pointIndex+1]]
        set x [expr $x+$dx]
        set y [expr $y+$dy]
        $data(canvas) coords $cObj [logCoord2primCoord $obj [lreplace $coords [expr 2*$pointIndex] [expr 2*$pointIndex+1] $x $y]]
    }
}

proc activeCanvas::getGroupPosition {obj group} {
    
    upvar 0 ::activeCanvas::$obj data
    set sumX 0
    set sumY 0
    set cnt 0
    foreach cObj [$data(canvas) find withtag group-$group] {
        
        
        set coords [$data(canvas) coords $cObj]
        foreach {x y} $coords {
            set sumX [expr $sumX + $x]
            set sumY [expr $sumY + $y]
            incr cnt
        }
        
    }
    if {$cnt == 0} { set cnt 1 }
    set result "[expr $sumX / $cnt] [expr $sumY / $cnt]"  
    return $result

}

proc activeCanvas::startShowingLinePreview {obj pos1 pos2} {
    upvar 0 ::activeCanvas::$obj data
    set data(previewLine) [$data(canvas) create line "$pos1 $pos2"]
}

proc activeCanvas::showLinePreview {obj pos1 pos2} {
    upvar 0 ::activeCanvas::$obj data
    if [info exists data(previewLine)] {
        $data(canvas) coords $data(previewLine) "$pos1 $pos2"
    } else {
        log::log error "startShowingLinePreview must called before showLinePreview"
    }
}

proc activeCanvas::endShowingLinePreview {obj} {
    upvar 0 ::activeCanvas::$obj data
    $data(canvas) delete $data(previewLine)
    unset data(previewLine)
}


proc activeCanvas::onVisualDrop_nothing {obj group startPos pos delta} {
}


proc activeCanvas::acceptDropFkt_all {canvas fromGroup toGroup} {
    return 1
}

proc activeCanvas::acceptDropFkt_none {canvas fromGroup toGroup} {
    return 0
}

proc activeCanvas::onDrop_nothing {canvas fromGroup toGroup absPos delta} {
}

proc activeCanvas::onContextMenu_nothing {group X Y} {
}

proc activeCanvas::onDoubleClick_nothing {group} {
}

proc activeCanvas::onClick_nothing {clickAttr} {

}


########################################################################################## interface


# PUBLIC
proc activeCanvas::create {obj} {
    # Creates a virtual widget with path $obj
    
    upvar 0 ::activeCanvas::$obj data  
    upvar 0 ::activeCanvas::persData_$obj persData 

    
    frame $obj
    canvas $obj.workspace -background white
    CDlg::mkScrollableXY $obj workspace

    set data(canvas) $obj.workspace
    set data(selgroup) ""
    set persData(zoom) 1.0


    startRedraw $obj
    bind $data(canvas) <ButtonPress-1> [ list activeCanvas::on_buttonPress1 $obj "%x %y" ]
    bind $data(canvas) <Control-ButtonPress-1> [ list activeCanvas::on_buttonPressAndControllKey $obj "%x %y" ]
    bind all <ButtonPress-1> +[list activeCanvas::on_buttonPressOutside $obj]
    bind $data(canvas) <Double-ButtonPress-1> [list activeCanvas::on_doubleClick $obj "%x %y"]
    bind $data(canvas) <ButtonPress-3> [list activeCanvas::on_buttonPress3 $obj "%x %y" %X %Y]
    bind $data(canvas) <Motion> [list activeCanvas::on_b1Motion $obj "%x %y" "%X %Y"]
    bind $data(canvas) <ButtonRelease-1> [list activeCanvas::on_buttonRelease $obj]
           
    set ::activeCanvas::activeCanvasId($data(canvas)) $obj

}

proc activeCanvas::delete {obj} {
    # bind all entfernen
    set binds [bind all <ButtonPress-1>]
    set search [list activeCanvas::on_buttonPressOutside $obj]
    set first [string first $search $binds]
    while {$first != -1} {
        set binds [string replace $binds $first [expr [string length $search] + $first -1]]
        set first [string first $search $binds]
    }
    bind all <ButtonPress-1> $binds
}


xn::proc activeCanvas::newElement {obj group type pos -normal "" -select "" -highlight "" -level 0 -rotate 0 -print 0} {
    # Creates a new canvas object of type $type at $pos and associates it to $group.
    # Args: -normal, -select and -highlight define the graphical properties
    # in the three states. Initially an element is in normal state. When it is selected it is in selected state.
    # When it is chosen as drop-target it is highlight state.
    # Beside the properties of the canvas item there are two special properties:
    # -clickable: Only clickable objects can be selected or draged by the mouse pointer.
    # -dropable: Only dropable objects can accept drops (just an optimization)

    upvar 0 ::activeCanvas::$obj data
    if { ![info exists data(onDrop,$group)] } {
        # default group definition, in case it is not provided by the user
        defGroup $obj $group
    }
    set res [createObject $obj $group $type $pos -normal $normal -select $select -highlight $highlight -level $level -rotate $rotate \
      -tags "print-$print"]
    return $res
}


proc activeCanvas::addAnimation {obj cObj animPath} {
    upvar 0 ::activeCanvas::$obj data
    lappend data(animationList) $cObj $animPath
}


xn::proc activeCanvas::defGroup {obj group \
      -acceptDropFkt "activeCanvas::acceptDropFkt_all" -onDrop "activeCanvas::onDrop_nothing" \
      -dominatorGroup "" -onVisualDrop "activeCanvas::onVisualDrop_nothing" \
      -onContextMenu activeCanvas::onContextMenu_nothing -onDoubleClick activeCanvas::onDoubleClick_nothing \
      -onClick activeCanvas::onClick_nothing} {

    # Defines the properties of a group. 
    # -acceptDropFkt: Function that determines which drop-targets are accepted. 
    # -onDrop: Function which is called when something is droped at this group
    # -dominatorGroup: When this group is selected it transfers the the selection to its dominator
    #    When the dominator is selected, all dominated groups are in selected mode.
    # -onVisualDrop: Function, that is executed when this group is highlighted for a drop
    # -onContextMenu: Function, that is executed when the right mouse button is pressed
    # -onDoubleClick: Function, that is executed when the left mouse button is double clicked

    upvar 0 ::activeCanvas::$obj data
    
    set data(defined,$group) 1
    set data(onDrop,$group) $onDrop
    set data(acceptDropFkt,$group) $acceptDropFkt
    set data(onVisualDrop,$group) $onVisualDrop
    set data(onContextMenu,$group) $onContextMenu
    set data(onDoubleClick,$group) $onDoubleClick
    set data(dominatorGroup,$group) $dominatorGroup
    set data(onClick,$group) $onClick
    if { $dominatorGroup ne "" } {
        lappend data(subgroups,$dominatorGroup) $group
    }
}

# PUBLIC 
proc activeCanvas::setAttr {obj group attr value} {
    upvar 0 ::activeCanvas::$obj data
    set data(attr-$attr,$group) $value
}

# PUBLIC 
proc activeCanvas::getAttr {obj group attr args} {
    upvar 0 ::activeCanvas::$obj data
    handleArgs [list -dest ""]
    if { ![empty $arg(dest)] } {
        upvar 1 $arg(dest) dest
        if { [info exists data(attr-$attr,$group)] } {
            set dest $data(attr-$attr,$group)
            return 1
        } else {
            return 0
        }
    } else {            
        return $data(attr-$attr,$group)
    }
}


# PUBLIC
proc activeCanvas::containsGroup {obj group} {
    upvar 0 ::activeCanvas::$obj data
    return [info exists data(defined,$group)]
}

# PUBLIC
proc activeCanvas::trySetSelectionAndScroll {obj group} {
    if { [containsGroup $obj $group] } {
        setSelectionAndScroll $obj $group
    }
}


# PUBLIC
proc activeCanvas::setSelectionAndScroll {obj group} {
    ensureVisible $obj $group
    setSelection $obj $group
}


# PUBLIC
proc activeCanvas::setSelectionWithEffect {obj group} {
    ensureVisible $obj $group
    setSelection $obj ""
    after 50 [list activeCanvas::setSelection $obj $group]
    after 100 [list activeCanvas::setSelection $obj ""]
    after 150 [list activeCanvas::setSelection $obj $group]
}


# PUBLIC
proc activeCanvas::setSelection {obj group {multi 0}} {
    upvar 0 ::activeCanvas::$obj data
    
    if {!$multi} {
        foreach old [getGroupsWithState $obj select] {
            setGroupState $obj $old normal
        }
    }

    if { ![empty $group] } {
        if { ![empty $data(dominatorGroup,$group)] } {
            set group $data(dominatorGroup,$group)
        }
    }
    set delSelect 0
    
    if {$multi} {
        set grpPos [lsearch $data(selgroups) $group]
        if [expr $grpPos == -1] {
            lappend data(selgroups) $group
            
        } else {
            set tmp $data(selgroups)
            set data(selgroups) [lrange $tmp 0 [expr $grpPos -1]]
            append data(selgroups) " " [lrange $tmp [expr $grpPos + 1] end]
            set delSelect 1
            #setGroupState $obj $group normal
        }
    } else {
        set data(selgroups) ""
        lappend data(selgroups) $group
    }
    
    #log::log XXX "Selected Groups: $data(selgroups)"
    
    if {$delSelect == 0} {
        set data(selgroup) $group
    }

    if { ![empty $group] } {
        if {$delSelect == 0} {
            setGroupState $obj $group select
        } else {
            setGroupState $obj $group normal
        }
        if { [info exists data(subgroups,$group)] } {
            foreach x $data(subgroups,$group) {
                if {$delSelect == 0} {
                    setGroupState $obj $x select
                } else {
                    setGroupState $obj $x normal
                }
            }
        }
    }
}

# PUBLIC
proc activeCanvas::getSelection {obj} {
    upvar 0 ::activeCanvas::$obj data
    return $data(selgroup)
}

proc activeCanvas::getMultiselection {obj} {
    upvar 0 ::activeCanvas::$obj data
    return $data(selgroups)
}

# PUBLIC
xn::proc activeCanvas::getBoundingBox {obj elemList} {
    upvar 0 ::activeCanvas::$obj data
    set newBbox [eval $data(canvas) bbox $elemList]
    >> xn::assert {![empty $newBbox]} "Bounding box empty, possible use of invisible color?"
    return [primCoord2logCoord  $obj $newBbox]
}


proc activeCanvas::ensureVisible {obj node} {
    upvar 0 ::activeCanvas::$obj data
 
    set bbox [eval $data(canvas) bbox group-$node]    
    set region [$data(canvas) cget -scrollregion]
  
    # fr x
    set mid [expr ([lindex $bbox 0] + [lindex $bbox 2])/2]
    set visfract [expr [lindex [$data(canvas) xview] 1] - [lindex [$data(canvas) xview] 0]]
    set fract [expr double($mid - [lindex $region 0]) / ([lindex $region 2] - [lindex $region 0]) - $visfract/2.0]
    $data(canvas) xview moveto $fract

    # fr y
    set mid [expr ([lindex $bbox 1] + [lindex $bbox 3])/2]
    set visfract [expr [lindex [$data(canvas) yview] 1] - [lindex [$data(canvas) yview] 0]]
    set fract [expr double($mid - [lindex $region 1]) / ([lindex $region 3] - [lindex $region 1]) - $visfract/2.0]
    $data(canvas) yview moveto $fract
}


# PUBLIC
proc activeCanvas::selectAfterUpdate {obj group} {
    upvar 0 ::activeCanvas::$obj data
    set data(selectAfterUpdate) $group
}


# PUBLIC
proc activeCanvas::startRedraw {obj} {
    upvar 0 ::activeCanvas::$obj data
    
    # delete window primitives
    if {[info exists data(winObjectGrp)]} {
        foreach windPrim $data(winObjectGrp) {
            destroy $windPrim
        }
    }
    if { [info exists data(selectAfterUpdate)] } {
        set lastSelected $data(selectAfterUpdate)
    } else {
        set lastSelected $data(selgroup)
    }
    catch {unset data}
    set data(animationList) ""
    set data(canvas) $obj.workspace
    set data(selgroup) ""
    set data(selgroups) ""
    set data(interactionState) "objectSelection"
    set data(oldObjects) [$data(canvas) find all]
    set data(lastSelected) $lastSelected
    set data(selectedDropPointSet) ""
    set data(dropPointRotationTimer) ""
    set data(clickedGroup) ""
    set data(toolbarCallback) ""
    set data(pressedInside) 0
    set data(countWindPrim) 0
    #xxx1
}


# PUBLIC 
proc activeCanvas::getBBox {obj} {
    upvar 0 ::activeCanvas::$obj data
    eval $data(canvas) delete $data(oldObjects)
    set data(oldObjects) ""
    set bbox [$data(canvas) bbox all]
    return $bbox
}



# PUBLIC 
proc activeCanvas::endRedraw {obj} {
    # Should be called after all objects are created. The function computes the bounding box
    # for the scrollbars.

    upvar 0 ::activeCanvas::$obj data
    set bbox [getBBox $obj]

    if { ![empty $bbox] } {
        # in case canvas is not empty: compute scrollregion
        assign "x1 y1 x2 y2" $bbox
        $data(canvas) config -scrollregion "0 0 [expr $x2+10] [expr $y2+10]"
    }
    $data(canvas) raise level-1
    $data(canvas) raise level-2

    if { ![empty $data(lastSelected)] && [containsGroup $obj $data(lastSelected)] } {
        setSelection $obj $data(lastSelected)
    }
    animate $obj
}


proc activeCanvas::computeAnimCoords {animPath animTime} {
    set keyTime0 -10
    set keyPos0 "0 0"
    foreach {keyTime keyPos} $animPath {
        if { $animTime<=$keyTime } {
            assign {keyx0 keyy0} $keyPos0
            assign {keyx  keyy } $keyPos
            set fraction [expr { double($animTime - $keyTime0) / ($keyTime - $keyTime0) }]
            set resultX [expr { round((1-$fraction)*$keyx0 + $fraction*$keyx) }]
            set resultY [expr { round((1-$fraction)*$keyy0 + $fraction*$keyy) }]
            return "$resultX $resultY"
        }
        set keyTime0 $keyTime
        set keyPos0 $keyPos
    }
    error "Invalid animPath"
}


proc activeCanvas::animate {obj} {
    upvar 0 ::activeCanvas::$obj data
    if { [empty $data(animationList)] } return

    set numSteps 60
    set delay 20

    for {set step 0} {$step <= $numSteps} {incr step} {
        set time [expr 100* $step / $numSteps]
        foreach {cObj animPath} $data(animationList) {
            $data(canvas) coords $cObj [computeAnimCoords $animPath $time]
        }
        update idletasks
        after $delay
    }
}


# PUBLIC 
proc activeCanvas::saveAsPostscript {obj filename} {
    upvar 0 ::activeCanvas::$obj data
    set bbox [$data(canvas) bbox print-1]
    if {[empty $bbox]} {
        puts "Nothing to print for file '$filename'"
        set bbox "0 0 0 0"
    }
    $data(canvas) postscript -x [lindex $bbox 0] -y [lindex $bbox 1] \
            -width [expr [lindex $bbox 2] - [lindex $bbox 0]] \
            -height [expr [lindex $bbox 3] - [lindex $bbox 1]] \
            -file $filename 
}

