
namespace eval basiccomp {
}

# Computes unique names for objNodes.
# Checks for uniqueness of given names.
# Supports namespaces


# Namensvarianten:
#
# SimpleName:
#    GivenName (falls nicht-leer) oder interne Objekt-Id
#   Bsp: add
#
# ShortName:
#   Bsp: FKT-add
#
# FullName:
#   Bsp: ::library::FKT-add
#
# RelativeName:
#   Namespaces relativ zu Bezugsposition abgeschnitten
#
#
# Der Parameter $useIndex bestimmt jeweils, ob der
# Name durch eine zustzlich angehngte Zahl (z.B. [2]) eindeutig
# gemacht werden soll.


# --- Dateninterface ---------------------------------------------------------

namespace eval names {}


# Konfigurations-Arrays:
#
#  names::isNamespace:           Classname -> 1
#  names::prefix:                Classname -> Prefix
#  names::initialNamePrefix      Classname -> Prefix
#  names::duplicatesAllowed      Classname -> 1


set names::isNamespace(Root) 1


# PRIVATE
proc basiccomp::getDuplicatesAllowed {objNode} {
    foreach class [c::getContexts $objNode] {
        if [info exists names::duplicatesAllowed($class)] {
            return $names::duplicatesAllowed($class)
        }
    }
    return 0
}


# PRIVATE
proc basiccomp::getPrefix {objNode} {
    # liefert namespace-prefix des Knotens
    foreach class [c::getContexts $objNode] {
        if [info exists names::prefix($class)] {
            return "$names::prefix($class)-"
        }
    }
    return ""
}



# PRIVATE
proc basiccomp::getNamespaceNode {node} {
    # returns a node, that represents the object's namespace
    # if $node is a nullptr, the nullptr is returned
    # the representation of Root's namespace is Root.

    if { [nullptr $node] || [c::isRoot $node] } {
        return $node
    }
    
    while 1 {
        set node [c::getParent $node]
        if [c::isLanguageObject $node] {
            foreach class [c::getContexts $node] {
                if [info exists names::isNamespace($class)] {
                    return $node
                }
            }
        }
        if [c::isRoot $node] {
            log::error [format "getNamespaceNode: %s is Root object!" [c::getContext $node]]
            return 0x0
        }
    }
}


# PRIVATE
proc basiccomp::getInitialNamePrefix {objNode} {
    foreach context [c::getContexts $objNode] {
        if { [info exists edithook::initValue($context.name)] } {
            return $edithook::initValue($context.name)
        } 
        if { [info exists names::initialNamePrefix($context)] } {
            return $names::initialNamePrefix($context)
        }
    }
    return [c::getClassName $objNode]
}



# --- Berechnung der Namenesvarianten ---------------------------------------------------


# HOOK
proc basiccomp::getGivenName {objNode} {
    # Liefert standardmig den Wert des Name-Attributs
    if [c::getSubElement $objNode name -var nameObj] {
        return [c::getValue $nameObj]
    } else {
        return ""
    }
}


# PUBLIC
proc basiccomp::getSimpleName {objNode {useIndex 1}} {
    if [nullptr $objNode] {
        return ""
    }
    set name [getGivenName $objNode]
    if [empty $name] {
        set name "obj[c::getPersistentId [d::getFile $objNode] $objNode]"
    }
    return "$name[getIndex $objNode $useIndex]"
}


# PRIVATE
proc basiccomp::getIndex {objNode {useIndex 1}} {
    if { [nullptr $objNode] || !$useIndex } {
        return ""
    }
    objData [c::getRoot $objNode]
    set name [getShortName $objNode 0]
    set namespace [getNamespaceNode $objNode]
    if { [info exists data(name,$namespace-$name)] && [llength $data(name,$namespace-$name)]>1 } {
        set index [expr [lsearch $data(name,$namespace-$name) $objNode]+1]
        return "\[$index\]"
    } else {
        return ""
    }
}


# PUBLIC
proc basiccomp::getShortName {objNode {useIndex 1}} {
    if [nullptr $objNode] {
        return ""
    }
    return "[getPrefix $objNode][getSimpleName $objNode $useIndex]"
}


# PUBLIC
proc basiccomp::getFullName {objNode {useIndex 1}} {
    if [nullptr $objNode] {
        return ""
    }
    return "[getNamespaceName $objNode $useIndex]::[getShortName $objNode $useIndex]"
}


# PRIVATE
proc basiccomp::getNamespaceName {objNode {useIndex 1}} {
    # Liefert den Namespace von $objNode
    # Beispiele: "", ::myModule::myFunction"
    if { [nullptr $objNode] } {
        return ""
    }
    set namespaceNode [getNamespaceNode $objNode]
    if { [nullptr $namespaceNode] || [c::isRoot $namespaceNode] } {
        return ""
    } else {
        return [getFullName $namespaceNode $useIndex]
    }
}


# PUBLIC
proc basiccomp::getRelativeName {node base {useIndex 1}} {
    if [nullptr $node] {
        return ""
    }
    set nsNode [getNamespaceNode $node]
    set nsBase [getNamespaceNode $base]

    if { [equal $nsNode $nsBase] || [c::isAncestorOf $nsBase $nsNode] } {
        set cut [string length "[getNamespaceName $node $useIndex]::"]
    } else {
        set cut 0
    }
    return [string range [getFullName $node $useIndex] $cut end]
}


# --- Suchen nach Namen --------------------------------------------------------------



proc basiccomp::findByFullName {root fullName args} {
    handleArgs "-var {}"
    upvar $arg(var) var

    set currObject $root
    set partList [splitFullName $fullName]
    foreach namePart [lrange $partList 1 end] {
        set new [findObject $currObject $namePart] 
        if [empty $new] {
            return 0
        }
        set currObject $new
    }
    set var $currObject
    return 1
}


proc basiccomp::splitFullName {s} {
    set result ""
    while 1 {
        set index [string first "::" $s]
        if { $index == -1 } {
            lappend result $s
            return $result
        } else {
            lappend result [string range $s 0 [expr $index-1]]
            set s [string range $s [expr $index+2] end]
        }
    }
}
        


proc basiccomp::findObject {namespace shortName} {
    objData [c::getRoot $namespace]

    if [info exists data(name,$namespace-$shortName)] {
        return $data(name,$namespace-$shortName)
    } else {
        return ""
    }
}



# --- Berechnung von Initialnamen ----------------------------------------------------


# PUBLIC
proc basiccomp::computeInitialName {objNode} {
    objData [c::getRoot $objNode]
    
    set namespace [getNamespaceNode $objNode]
    set prefix [getPrefix $objNode]
    set initialNamePrefix [getInitialNamePrefix $objNode]
    set i 1

    while 1 {
        set name "$initialNamePrefix$i"
        set namespace [getNamespaceNode $objNode]
        if ![info exists data(name,$namespace-${prefix}$name)] {
            return $name
        }
        incr i
    }
}


# --- Verschiedenes --------------------------------------------------------------



# PUBLIC
proc basiccomp::getErrorMessages {root messageListName} {
    upvar $messageListName messageList
    objData $root
    foreach {name objList} [array get data name,*] {
        if { [llength $objList]>1 } {
            foreach obj $objList {
                if { ![getDuplicatesAllowed $obj] } {
                    checkutil::appendErrorItem messageList $obj [i18n::text "Name is not unique: %s" [getFullName $obj]]
                }
            }
        }
    }
}


# PUBLIC (called by "file" after update)
proc basiccomp::run {root} {
    objData $root
    catch {unset data}

    c::traverseTree $root isLanguageObject objNode {
        lappend data(name,[getNamespaceNode $objNode]-[getShortName $objNode 0]) $objNode
    }
}

