
namespace eval ::checkutil {}

proc checkutil::checkProgram {obj} {
    set list ""
    basiccomp::getErrorMessages $obj list
    checkProgram_i $obj list
    return $list
}

proc checkutil::checkProgram_i {obj listName} {
    # checks subtree $obj and appends results to $listName
    upvar $listName list
    if { [c::isLanguageObject $obj] } {
        set result [doNodeCheck $obj]
    } elseif { [equal [c::getClassName $obj] ObservedType] } {
        set value [c::getValue $obj]
        set result [doValueCheck $obj $value]
    } else {
        set result [doAggrCheck $obj]
    }
    appendErrorItem list $obj $result
    foreach sub [c::getSubElements $obj] {
        checkProgram_i $sub list
    }
}


proc checkutil::appendErrorItem {listName obj msgText} {
    if { ![empty $msgText] } {
        upvar $listName list
        lappend list [list $obj orange "ERROR: [c::getContext $obj]: $msgText"]
    }
}


proc checkutil::checkType_VLString {x} {
    return 1
}

proc checkutil::checkType_VLPointer {x} {
    return 1
}

proc checkutil::checkType_VLBoolean {x} {
    return [elementOf $x "0 1"]
}

proc checkutil::checkType_VLInt {x} {
    return [regexp {^-?(0|[1-9][0-9]{0,8})$} $x]
}

proc checkutil::checkType_VLDouble {x} {
    return [regexp {^-?(0|[1-9][0-9]{0,8})(\.[0-9]+)?$} $x]
}


proc checkutil::checkType_VLPoint {x} {
    if { [llength $x]!=2 } { return 0}
    foreach elem $x {
        if ![checkType_VLDouble $elem] {return 0}
    }
    return 1
}

proc checkutil::checkType_VLRegion {x} {
    if { ![llength $x]==4 } { return 0}
    foreach elem $x {
        if ![checkType_VLDouble $elem] {return 0}
    }
    return 1
}


proc checkutil::checkType_VLPointArray {x} {
    if { ([llength $x]%2)!=0 } { return 0}
    foreach elem $x {
        if ![checkType_VLDouble $elem] {return 0}
    }
    return 1
}


proc checkutil::checkType_VLList {x} {
    return [expr {![catch {llength $x}]}]
}


proc checkutil::doNodeCheck {obj} {
    variable data

    foreach context [c::getContexts $obj] {
        if [info exists data(check-$context)] {
            foreach proc $data(check-$context) {
                set resultCode [catch [lindex $proc 1] resultValue]
                if { $resultCode == 1 } {
                    log::log error "Internal error while checking node $obj"
                    puts $::errorInfo
                    return "Internal error while checking node $obj"
                }
                if { $resultCode != 2 } {
                    log::log error "Internal error while checking node $obj: Return missing?"
                    return "Internal error while checking node $obj: Return missing?"
                }
                if ![empty $resultValue] {
                    return $resultValue
                }
            }
        }
    }
    return ""
}


proc checkutil::doValueCheck {obj value} {
    # Exectuted for VAL and REF Nodes
    variable data

    if { ([c::isValue $obj] && [empty $value]) || ([c::isReference $obj] && [nullptr $value]) } {
        if { [c::getNullAllowed $obj] } {
            return ""
        } else {
            if [c::isValue $obj] {
                return "Value must not be empty"
            } else {
                return "Reference must not be null."
            }
        }
    }

    if ![checkType_[c::getTypeName $obj] $value] {
        return "Not a valid instance of type [c::getTypeName $obj]."
    }
    foreach context [c::getContexts $obj] {        
        if [info exists data(check-$context)] {
            foreach proc $data(check-$context) {
                set resultCode [catch [lindex $proc 1] resultValue]
                if { $resultCode == 1 } {
                    log::error "Internal error while checking value '$value' in context '$obj': $resultValue"
                    return "Internal error"
                }
                if { $resultCode != 2 } {
                    log::error "Internal error while checking value '$value' in context '$obj': Return missing?"
                    return "Internal error"
                }
                if ![empty $resultValue] {
                    return $resultValue
                }
            }
        }
    }
    return ""
}


proc checkutil::doAggrCheck {obj} {
    variable data
    if { [c::minProcessorMultiplicityIsOne $obj] && [c::getSubElementCount $obj]==0 } {
        return [i18n::text "Required sub-element is missing"]
    }
    foreach context [c::getContexts $obj] {        
        if [info exists data(check-$context)] {
            foreach proc $data(check-$context) {
                set resultCode [catch [lindex $proc 1] resultValue]
                if { $resultCode == 1 } {
                    return "Internal error while checking aggregate: $resultValue"
                }
                if { $resultCode != 2 } {
                    return "Internal error while checking aggregate: Return missing?"
                }
                if ![empty $resultValue] {
                    return $resultValue
                }
            }
        }
    }
    return ""
}


proc checkutil::addCheck {elementName proc {level 2}} {
    # proc's can access two variables:
    # $obj: contains the obj id
    # $value: contains the new value, which must be checked

    variable data
    lappend data(check-$elementName) [list $level $proc]
    set data(check-$elementName) [lsort -index 0 -integer $data(check-$elementName)]
}


# usable functions

proc checkutil::checkRange {min max} {
    upvar obj obj
    upvar value value
    if { $value < $min || $value > $max } {
        return "Value must be in range $min..$max"
    }
    return ""
}

proc checkutil::checkNoBlanks {} {
    upvar obj obj
    upvar value value
    if [string first " " $value]!=-1 {
        return "String must not contain blanks"
    }
    return ""
}



proc checkutil::checkType {typename} {
    upvar obj obj
    upvar value value

    if [equal $value "0x0"] {
        return ""
    }
    if ![c::isSubtypeOf [c::getTypeId $value] [c::${typename}_type]] {
        return "Not a subtype of $typename!"
    }
    return ""
}


proc checkutil::checkUndefinedReference {typename} {
    upvar obj obj
    upvar value value

    if [equal $value "0x0"] {
        return "Undefined Reference"
    }
    return ""
}

