
# Module: file.tcl
#
# Reprsentiert einen Strukturbaum. Untersttzt das Laden und Speichern,
# verwaltet Sichten darauf und stoesst Aktualisierungen dieser Sichten an.

namespace eval file {
    set cdata(files) ""
    set cdata(lastUnnamedNumber) 0
    set cdata(processors) ""
}


### klassenfunktionen

proc file::findNodeByInternalId {} {
    if [CDlg::getString [i18n::text "Find node by internal ID"] node] {
        highlightNodeInTreeView $node
    }
}


proc file::getInitialDir {id} {
    variable initialDir
    if [info exists initialDir($id)] {
        return $initialDir($id)
    } else {
        return "."
    }
}

proc file::setInitialDir {id location} {
    variable initialDir
    set initialDir($id) $location
}


proc file::registerProcessor {name cmd {label ""}} {
    variable cdata
    if {[empty $label]} {set label $name}
    lappend cdata(processors) $name
    set cdata(processor,$name) $cmd
    set cdata(processorLabel,$name) $label
}


proc file::getProcessorLabel {name} {
    variable cdata
    return $cdata(processorLabel,$name)
}


proc file::getProcessorNames {} {
    variable cdata
    return $cdata(processors)
}


proc file::registerFile {oid} {
    variable cdata
    lappend cdata(files) $oid
}

proc file::unregisterFile {oid} {
    variable cdata
    lremove cdata(files) $oid
}

proc file::getAllFiles {} {
    variable cdata
    return $cdata(files)
}


proc file::highlightNodeInTreeView {node} {
    if [nullptr $node] {
     ::CDlg::Error "Null Pointer selected!"
     return
    }
    set oid [getFileOfNode $node]
    oop::data $oid
    set view [openView $oid [getRoot $oid] "Tree view"]
    $view highlightNode $node
}


proc file::highlightNode {node} {
    set oid [getFileOfNode $node]
    oop::data $oid
    
    # find suitable view node
    set viewNode $node
    while { ![nullptr $viewNode] } {
        set list [treedraw::getViewNames $viewNode]
        if ![empty $list] {
            set viewType [lindex $list 0]
            break
        } else {
            set viewNode [c::getParent $viewNode]
        }
    }

    if { ![nullptr $viewNode] } {
        set view [openView $oid $viewNode $viewType]
        $view selectNode $node
    } else {
        highlightNodeInTreeView $node
    }
}


proc file::getFileOfRoot {root} {
    variable cdata
    return $cdata(fileOfRoot-$root)
}

proc file::getFileOfNode {node} {
    return [getFileOfRoot [c::getRoot $node]]
}

proc file::getFiletypes {} {
    set result ""
    lappend result [list "Editor Data" $::cfg::extension]
    lappend result [list "All Files" "*"]
    return $result
}

proc file::getDefaultExtension {} {
    return $::cfg::extension
}


#### Konstruktoren


proc file::new {} {
    set oid [oop::new]
    setRoot $oid [xmlio::getNewFile]
    change::initUndoList [getRoot $oid]
    registerFile $oid
    return $oid
}

proc file::openNoWindow {filename} {
    set oid [oop::new]
    setRoot $oid [xmlio::load $filename]
    change::initUndoList [getRoot $oid]
    setFilename $oid $filename
    registerFile $oid
    return $oid
}


proc file::openWithWindow {filename} {
    set oid [openNoWindow $filename]
    if ![empty $oid] {
        openMainView $oid
    }
    return $oid
}


proc file::interactiveOpen {} {
    
    set filename [tk_getOpenFile -filetypes [getFiletypes] -defaultextension [getDefaultExtension] -initialdir [getInitialDir "xmlFiles"]] 
    if {$filename == ""} return ""
    setInitialDir "xmlFiles" [file dirname $filename]
    return [openWithWindow $filename]
}

### objektfunktionen

proc file::undo {oid} {
    if ![change::undo [getRoot $oid]] {
        CDlg::Information [i18n::text "Undo is not available."] "Undo"
    }
}


proc file::openAbstractView {oid} {
    openView $oid [getRoot $oid] "Tree view"
}


proc file::openErrorView {oid} {
    openView $oid [getRoot $oid] "Check view"
}


proc file::openMainView {oid} {
    set list [treedraw::getViewNames [getRoot $oid]]
    if [empty $list] {
        openView $oid [getRoot $oid] "Tree view"
    } else {
        openView $oid [getRoot $oid] [lindex $list 0]

    }
}


proc file::openAttributeDialog {oid viewNode} {
    set className [c::getClassName $viewNode]

    if [info exists ::treedraw::specialDialog($className)] {
        set viewType $::treedraw::specialDialog($className)
        set view [xn::new SpecialDialogView $viewNode $viewType]
    } else {
        treedraw::openStdAttributeDialog $viewNode
    }
}


proc file::openView {oid viewNode viewType args} {
    oop::data $oid
    handleArgs "-newView 0"

    if { [MDIFrame::getReuseViews] && !$arg(newView) } {
        foreach abstractView $data(views) {
            if [$abstractView isReusable $viewNode $viewType] {
                $abstractView makeVisible
                return $abstractView
            }
        }
    }
    switch $viewType {
        "Tree view" {
            set view [xn::new TreeView $viewNode]
        }
        "Check view" {
            set view [xn::new CheckView $viewNode]
        }
        default {
            set view [xn::new LidoView $viewNode $viewType]
        }
    }
    #updateNow $oid
    return $view
}


proc file::beforeDeletion {oid node} {
    oop::data $oid
    foreach abstractView $data(views) {
        $abstractView beforeDeletion $node
    }
}

proc file::saveAs {oid filename} {
    xmlio::save [getRoot $oid] $filename
    setFilename $oid $filename
    setModified $oid 0
}


proc file::interactiveSave {oid} {
    # returns: 1, if the file was saved

    set filename [getFilename $oid]
    if [getUnnamed $oid] {
        return [interactiveSaveAs $oid]
    } else {
        saveAs $oid $filename
        return 1
    }
}


proc file::interactiveSaveAs {oid} {
    # returns: 1, if the file was saved

    set filename [tk_getSaveFile -filetypes [getFiletypes] -defaultextension [getDefaultExtension] -initialdir [getInitialDir "xmlFiles"]]
    if {$filename == ""} {return 0}
    setInitialDir "xmlFiles" [file dirname $filename]
    saveAs $oid $filename
    return 1

}


proc file::getFilenameNoPath {oid} {
    return [file tail [getFilename $oid]]
}


proc file::getFilename {oid} {
    # liefert Dateiename mit komplettem Pfad.

    oop::data $oid
    return $data(filename)
}


proc file::setFilename {oid filename} {
    oop::data $oid
    set data(filename) [file join [pwd] $filename]
    set data(unnamed) 0
    forceUpdate $oid
}


proc file::setReadonly {oid value} {
    oop::data $oid
    set data(readonly) $value
}


proc file::getReadonly {oid} {
    oop::data $oid
    return $data(readonly)
}


proc file::getUnnamed {oid} {
    oop::data $oid
    return $data(unnamed)
}

proc file::setModified {oid value} {
    oop::data $oid
    set data(modified) $value
    forceUpdate $oid
}

proc file::getModified {oid} {
    oop::data $oid
    return $data(modified)
}

proc file::setRoot {oid root} {
    # privat!
    variable cdata
    oop::data $oid

    set cdata(fileOfRoot-$root) $oid
    set data(root) $root
    set data(updateScheduled) 0
    set data(syncCount) 0
    set data(views) ""
    set data(modified) 0
    set data(unnamed) 1
    incr cdata(lastUnnamedNumber)
    set data(filename) [i18n::text "Unnamed%s" $cdata(lastUnnamedNumber)]
    set data(afterUpdate) ""
    set data(readonly) 0
    set data(eventManager) [eventManager::new]
    registerUpdateHandlers $oid
    forceUpdate $oid
}


proc file::getRoot {oid} {
    oop::data $oid
    return $data(root)
}

proc file::addView {oid abstractView} {
    oop::data $oid
    lappend data(views) $abstractView
    eventManager::addListenerObj $data(eventManager) $abstractView "" "" $::file::PRI_VIEWUPDATE [list $abstractView update]
    #updateLater $oid
}


proc file::getViews {oid} {
    oop::data $oid
    return $data(views)
}


proc file::removeView {oid abstractView} {
    oop::data $oid
    lremove data(views) $abstractView
    eventManager::removeListener $data(eventManager) $abstractView
}


proc file::startProcessor {oid processor} {
    if [equal $processor "EXPORT_HTML"] {
        exportRepresentation $oid html [main::getParam dest]
        return 1
    } elseif [equal $processor "EXPORT_EPS"] {
        exportRepresentation $oid eps [main::getParam dest]
        return 1
    } elseif [equal $processor "EXPORT_PNG"] {
        exportRepresentation $oid png [main::getParam dest]
        return 1
    } else {
        return [startLigaProcessor $oid $processor]
    }
}


proc file::startLigaProcessor {oid processor} {
    variable cdata
    if ![elementOf $processor [getProcessorNames]] {
        log::log error "There is no processor '$processor'"
        return 0
    }
    set check1 [checkutil::checkProgram [getRoot $oid]]
    if ![empty $check1] {
        openErrorView $oid
        return 0
    }
    log::process "PROCESS" "Running processor '$processor'"
    vlMessage::reset
    eval $cdata(processor,$processor) [getRoot $oid]
    if { [vlMessage::getErrorLevel] >= 0 } {
        xn::new MessageView [getRoot $oid] "Errors" [vlMessage::getList]
        set result 0
    } else {
        set result 1
    }
    log::process "" "Processor ready -------------------------------------------"
    return $result
}


proc file::findUsages {oid targetObject} {
    set list ""
    foreach ref [c::findReferences $targetObject [getRoot $oid]] {
        lappend list [list $ref green [c::getObjectRep $ref]]
    }
    xn::new MessageView [getRoot $oid] [i18n::text "Usages of %s" [basiccomp::getFullName $targetObject]] $list
}


proc file::initSearchData {} {
    variable search

    set search(string,value) ""
    set search(string,label) [i18n::text "Search pattern"]
    set search(string,editor) Entry
    set search(string,check) ""

    set search(substrings,value) 1
    set search(substrings,label) [i18n::text "Match substrings"]
    set search(substrings,editor) "Checkbox"
    set search(substrings,check) ""

    set search(regexp,value) 0
    set search(regexp,label) [i18n::text "Regular expression"]
    set search(regexp,editor) "Checkbox"
    set search(regexp,check) ""

    set search(caseSens,value) 0
    set search(caseSens,label) [i18n::text "Case sensitive"]
    set search(caseSens,editor) "Checkbox"
    set search(caseSens,check) ""

    set search(onlyNames,value) 0
    set search(onlyNames,label) [i18n::text "Only name attributes"]
    set search(onlyNames,editor) "Checkbox"
    set search(onlyNames,check) ""
}


proc file::findRegexp {oid} {
    variable search
    if {![info exists search]} {
        initSearchData
    }
    if [CDlg::stdInputDialog [i18n::text "Search for value"] [list string substrings regexp caseSens onlyNames] search] {

        # option processing
        if $search(caseSens,value) {
            set opt ""
        } else {
            set opt "-nocase"
        }
        if $search(regexp,value) {
            set pattern $search(string,value)
        } else {
            set pattern [quote::regexpQuote $search(string,value)]
        }
        if !$search(substrings,value) {
            set pattern "^${pattern}\$"
        }

        set list ""
        c::traverseTree [getRoot $oid] isValue obj {
            if { !$search(onlyNames,value) || [equal [c::getRole $obj] "name"] } {
                if [eval regexp $opt -- [list $pattern [c::getValue $obj]]] {
                    lappend list [list $obj green [c::getObjectRep $obj]]
                }
            }
        }
        xn::new MessageView [getRoot $oid] [i18n::text "Search for '%s'" $search(string,value)] $list
    }
}

proc ::file::close {oid} {
    foreach view [getViews $oid] {
        $view close
    }
    unregisterFile $oid
}

proc ::file::tryClose {oid} {
    if [confirmClose $oid] {
        close $oid
    }
}

proc ::file::confirmClose {oid} {
    # Returns 1, if the file should be closed, 0 otherwise.
                                                                                                            
    variable data
                                                                                                            
    if ![getModified $oid] {
        return 1
    }
                                                                                                            
    set msg ""
    append msg "[getFilename $oid]\n\n"
    append msg [i18n::text "File contains unsaved changes. Save it first?"]

    set res [CDlg::Question $msg "Close"]
    switch $res {
        yes {
            return [interactiveSave $oid]
        }
        no {
            return 1
        }
        cancel {
            return 0
        }
    }
}

proc ::file::tryCloseAll {} {
    # returns 1, if the application should be closed

    foreach file [getAllFiles] {
        if ![confirmClose $file] {
            return 0
        }
    }
    return 1
}



proc file::exportRepresentationInteractive {oid} {
    variable exportRepData
    if {![info exists exportRepData]} {
        set exportRepData(format,value) "html"
        set exportRepData(format,label) [i18n::text "Export format"]
        set exportRepData(format,editor) "Radio -values {html html eps eps png png}"
        set exportRepData(format,check) ""

        set exportRepData(path,value) "/tmp/export"
        set exportRepData(path,label) [i18n::text "Target directory"]
        set exportRepData(path,editor) "Path -filetypes dir"
        set exportRepData(path,check) ""
    }
    if [CDlg::stdInputDialog [i18n::text "Export representation"] [list format path] exportRepData] {
        exportRepresentation $oid $exportRepData(format,value) $exportRepData(path,value)
        CDlg::Information [i18n::text "Export done!"] [i18n::text "Export representation"]
    }
}


proc file::exportRepresentation {oid type destDir} {
    if ![file exists $destDir] {
        file mkdir $destDir
    }
    exportHtml::export $oid $destDir $type
}



proc file::showInfoDialog {oid} {
    lappend message [i18n::text "Filename: %s" [getFilename $oid]]
    lappend message [i18n::text "Tree nodes: %s" [c::getObjCount [getRoot $oid]]]
    lappend message [i18n::text "Modified: %s" [lindex [i18n::text "no yes"] [getModified $oid]]]

    CDlg::Information [join $message "\n\n"] [i18n::text "File Information"]
}


# --- update -----------------------------------------------------------------------


set ::file::PRI_CONTEXTSENSITIVEINIT 1
set ::file::PRI_SYNC 2
set ::file::PRI_BASICCOMP 3
set ::file::PRI_VIEWUPDATE 4
set ::file::PRI_AFTERUPDATE 5


proc file::registerUpdateHandlers {oid} {
    oop::data $oid
    eventManager::addListener $data(eventManager) "" "" $::file::PRI_CONTEXTSENSITIVEINIT [list edit::doContextSensitiveInit]
    eventManager::addListener $data(eventManager) "" "" $::file::PRI_SYNC [list synchook::syncProgram $data(root)]
    eventManager::addListener $data(eventManager) "" "" $::file::PRI_BASICCOMP [list basiccomp::run $data(root)]
    eventManager::addListener $data(eventManager) "" "" $::file::PRI_AFTERUPDATE [list file::runAfterUpdate $oid]
}


proc file::changeOccured {changedNode} {
    treePath::resetCache

    variable cdata
    set root [d::getFile $changedNode]
    if { [info exists cdata(fileOfRoot-$root)] } {
        set file [getFileOfRoot $root]
        changeOccured2 $file $changedNode
    }
}

proc file::changeOccured2 {oid changedNode} {
    oop::data $oid

    # Kein change bei copy-Operationen!
    if {[equal [c::getRoot $changedNode] [getRoot $oid]]} {
        eventManager::addChange $data(eventManager) $changedNode
        setModified $oid 1    
    }
}


proc file::runAfterUpdate {oid} {    
    oop::data $oid
    change::addCheckpoint $data(root)
    eval $data(afterUpdate)
    set data(afterUpdate) ""
}

proc file::afterUpdate {oid cmd} {
    oop::data $oid
    if { !$data(updateScheduled) } {
        eval $cmd
    } else {
        set data(afterUpdate) $cmd
    }
}

proc file::forceUpdate {oid} {
    oop::data $oid
    eventManager::addChange $data(eventManager) [getRoot $oid]
}

# ---- old functions


proc file::updateNow {oid} {
    forceUpdate $oid
}


